from __future__ import annotations

import inspect
import logging
import subprocess
import sys
import warnings
from contextlib import ExitStack
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Type

import pytest

from rdflib import RDF, SKOS
from rdflib.namespace import DefinedNamespace, Namespace
from rdflib.term import URIRef
from test.data import TEST_DATA_DIR


def test_definednamespace_creator_qb():
    """
    Tests basic use of DefinedNamespace creator script using QB
    """

    definednamespace_script = (
        Path(__file__).parent.parent.parent
        / "rdflib"
        / "tools"
        / "defined_namespace_creator.py"
    )
    qb_data_file = Path(TEST_DATA_DIR) / "defined_namespaces" / "qb.ttl"
    print("\n")
    print(f"Using {definednamespace_script}...")
    print(f"Testing {qb_data_file}...")
    completed = subprocess.run(
        [
            sys.executable,
            str(definednamespace_script),
            str(qb_data_file),
            "http://purl.org/linked-data/cube#",
            "QB",
        ],
        capture_output=True,
        text=True,
    )
    assert completed.returncode == 0, "subprocess exited incorrectly"
    assert Path.is_file(Path("_QB.py")), "_QB.py file not created"
    has_ns = False
    has_test_class = False
    with open(Path("_QB.py")) as f:
        for line in f.readlines():
            if '_NS = Namespace("http://purl.org/linked-data/cube#")' in line:
                has_ns = True
            if (
                "Attachable: URIRef  # Abstract superclass for everything that can have attributes and dimensions"
                in line
            ):
                has_test_class = True
    assert has_ns, "_QB.py does not contain _NS"
    assert has_test_class, "_QB.py does not class Attachable"

    # cleanup
    Path.unlink(Path("_QB.py"))


def test_definednamespace_creator_fake():
    """
    Tests incorrect use of DefinedNamespace creator script -
    RDF file of unknonwn type
    """

    definednamespace_script = (
        Path(__file__).parent.parent.parent
        / "rdflib"
        / "tools"
        / "defined_namespace_creator.py"
    )
    qb_data_file = Path(TEST_DATA_DIR) / "defined_namespaces" / "fake.xxx"
    print("\n")
    print(f"Using {definednamespace_script}...")
    print(f"Testing {qb_data_file}...(expected to fail)")
    completed = subprocess.run(
        [
            sys.executable,
            str(definednamespace_script),
            str(qb_data_file),
            "http://purl.org/linked-data/cube#",
            "QB",
        ],
        capture_output=True,
        text=True,
    )
    assert completed.returncode == 1, "subprocess exited incorrectly (failure expected)"


def test_definednamespace_creator_bad_ns():
    """
    Tests incorrect use of DefinedNamespace creator script -
    supplied namespace doesn't end in # or /
    """

    definednamespace_script = (
        Path(__file__).parent.parent.parent
        / "rdflib"
        / "tools"
        / "defined_namespace_creator.py"
    )
    qb_data_file = Path(TEST_DATA_DIR) / "defined_namespaces" / "fake.xxx"
    print("\n")
    print(f"Using {definednamespace_script}...")
    print(f"Testing {qb_data_file}...(expected to fail - bad NS given)")
    completed = subprocess.run(
        [
            sys.executable,
            str(definednamespace_script),
            str(qb_data_file),
            "http://purl.org/linked-data/cube",
            "QB",
        ],
        capture_output=True,
        text=True,
    )
    assert completed.returncode == 1, "subprocess exited incorrectly (failure expected)"


def test_definednamespace_dir():
    x = dir(RDF)

    values = [
        RDF.nil,
        RDF.direction,
        RDF.first,
        RDF.language,
        RDF.object,
        RDF.predicate,
        RDF.rest,
        RDF.subject,
        RDF.type,
        RDF.value,
        RDF.Alt,
        RDF.Bag,
        RDF.CompoundLiteral,
        RDF.List,
        RDF.Property,
        RDF.Seq,
        RDF.Statement,
        RDF.HTML,
        RDF.JSON,
        RDF.PlainLiteral,
        RDF.XMLLiteral,
        RDF.langString,
    ]

    assert len(values) == len(x)

    for value in values:
        assert value in x


def test_definednamespace_jsonld_context():
    expected = {
        "@context": {
            "skos": "http://www.w3.org/2004/02/skos/core#",
            "altLabel": "skos:altLabel",
            "broadMatch": "skos:broadMatch",
            "broader": "skos:broader",
            "broaderTransitive": "skos:broaderTransitive",
            "changeNote": "skos:changeNote",
            "closeMatch": "skos:closeMatch",
            "definition": "skos:definition",
            "editorialNote": "skos:editorialNote",
            "exactMatch": "skos:exactMatch",
            "example": "skos:example",
            "hasTopConcept": "skos:hasTopConcept",
            "hiddenLabel": "skos:hiddenLabel",
            "historyNote": "skos:historyNote",
            "inScheme": "skos:inScheme",
            "mappingRelation": "skos:mappingRelation",
            "member": "skos:member",
            "memberList": "skos:memberList",
            "narrowMatch": "skos:narrowMatch",
            "narrower": "skos:narrower",
            "narrowerTransitive": "skos:narrowerTransitive",
            "notation": "skos:notation",
            "note": "skos:note",
            "prefLabel": "skos:prefLabel",
            "related": "skos:related",
            "relatedMatch": "skos:relatedMatch",
            "scopeNote": "skos:scopeNote",
            "semanticRelation": "skos:semanticRelation",
            "topConceptOf": "skos:topConceptOf",
            "Collection": "skos:Collection",
            "Concept": "skos:Concept",
            "ConceptScheme": "skos:ConceptScheme",
            "OrderedCollection": "skos:OrderedCollection",
        }
    }
    actual = SKOS.as_jsonld_context("skos")

    assert actual == expected


prefix = "http://example.com/"


class DFNSNoNS(DefinedNamespace):
    defined: URIRef
    _defined: URIRef


class DFNSDefaults(DefinedNamespace):
    _NS = Namespace(f"{prefix}DFNSDefaults#")
    defined: URIRef
    _defined: URIRef


class DFNSDefaultsEmpty(DefinedNamespace):
    _NS = Namespace(f"{prefix}DFNSDefaultsEmpty#")


class DFNSWarnFailEmpty(DefinedNamespace):
    _NS = Namespace(f"{prefix}DFNSWarnFailEmpty#")
    _warn = True
    _fail = True


class DFNSNoWarnNoFail(DefinedNamespace):
    _NS = Namespace(f"{prefix}DFNSNoWarnNoFail#")
    _warn = False
    _fail = False
    defined: URIRef
    _defined: URIRef


class DFNSNoWarnFail(DefinedNamespace):
    _NS = Namespace(f"{prefix}DFNSNoWarnFail#")
    _warn = False
    _fail = True
    defined: URIRef
    _defined: URIRef


class DFNSWarnNoFail(DefinedNamespace):
    _NS = Namespace(f"{prefix}DFNSWarnNoFail#")
    _warn = True
    _fail = False
    defined: URIRef
    _defined: URIRef


class DFNSWarnFail(DefinedNamespace):
    _NS = Namespace(f"{prefix}DFNSWarnFail#")
    _warn = True
    _fail = True
    defined: URIRef
    _defined: URIRef


@dataclass
class DFNSInfo:
    dfns: Type[DefinedNamespace]
    suffix: Optional[str]
    has_attrs: bool = True


dfns_infos = [
    DFNSInfo(DFNSNoNS, None),
    DFNSInfo(DFNSDefaults, "DFNSDefaults#"),
    DFNSInfo(DFNSNoWarnNoFail, "DFNSNoWarnNoFail#"),
    DFNSInfo(DFNSWarnFail, "DFNSWarnFail#"),
    DFNSInfo(DFNSNoWarnFail, "DFNSNoWarnFail#"),
    DFNSInfo(DFNSWarnNoFail, "DFNSWarnNoFail#"),
    DFNSInfo(DFNSDefaultsEmpty, "DFNSDefaultsEmpty#", False),
    DFNSInfo(DFNSWarnFailEmpty, "DFNSWarnFailEmpty#", False),
    DFNSInfo(DefinedNamespace, None, False),
]
dfns_list = [item.dfns for item in dfns_infos]


def get_dfns_info(dfns: Type[DefinedNamespace]) -> DFNSInfo:
    for dfns_info in dfns_infos:
        if dfns_info.dfns is dfns:
            return dfns_info
    raise ValueError("No DFNSInfo for the DefinedNamespace passed in ...")


@pytest.fixture(
    scope="module",
    params=[item.dfns for item in dfns_infos],
)
def dfns(request) -> DFNSInfo:
    assert issubclass(request.param, DefinedNamespace)
    return request.param


def test_repr(dfns: Type[DefinedNamespace]) -> None:
    dfns_info = get_dfns_info(dfns)
    ns_uri = f"{prefix}{dfns_info.suffix}"
    logging.debug("ns_uri = %s", ns_uri)

    repr_str: Optional[str] = None

    with ExitStack() as xstack:
        if dfns_info.suffix is None:
            xstack.enter_context(pytest.raises(AttributeError))
        repr_str = f"{dfns_info.dfns!r}"
    if dfns_info.suffix is None:
        assert repr_str is None
    else:
        assert repr_str is not None
        repro = eval(repr_str)
        assert ns_uri == f"{repro}"


def test_inspect(dfns: Type[DefinedNamespace]) -> None:
    """
    `inspect.signature` returns. This is here to check that this does not
    trigger infinite recursion.
    """
    inspect.signature(dfns, follow_wrapped=True)


@pytest.mark.parametrize(
    ["attr_name", "is_defined"],
    [
        ("defined", True),
        ("_defined", True),
        ("notdefined", False),
        ("_notdefined", False),
    ],
)
def test_value(dfns: Type[DefinedNamespace], attr_name: str, is_defined: bool) -> None:
    dfns_info = get_dfns_info(dfns)
    if dfns_info.has_attrs is False:
        is_defined = False
    resolved: Optional[str] = None
    with ExitStack() as xstack:
        warnings_record = xstack.enter_context(warnings.catch_warnings(record=True))
        if dfns_info.suffix is None or (not is_defined and dfns._fail is True):
            xstack.enter_context(pytest.raises(AttributeError))
        resolved = eval(f"dfns.{attr_name}")
    if dfns_info.suffix is not None:
        if is_defined or dfns._fail is False:
            assert f"{prefix}{dfns_info.suffix}{attr_name}" == f"{resolved}"
        else:
            assert resolved is None
        if dfns._warn is False:
            assert len(warnings_record) == 0
        elif not is_defined and resolved is not None:
            assert len(warnings_record) == 1
    else:
        assert resolved is None


@pytest.mark.parametrize(
    ["attr_name", "is_defined"],
    [
        ("defined", True),
        ("_defined", True),
        ("notdefined", False),
        ("_notdefined", False),
    ],
)
def test_contains(
    dfns: Type[DefinedNamespace], attr_name: str, is_defined: bool
) -> None:
    dfns_info = get_dfns_info(dfns)
    if dfns_info.suffix is not None:
        logging.debug("dfns_info = %s", dfns_info)
    if dfns_info.has_attrs is False:
        is_defined = False
    does_contain: Optional[bool] = None
    with ExitStack() as xstack:
        if dfns_info.suffix is None:
            xstack.enter_context(pytest.raises(AttributeError))
        does_contain = attr_name in dfns
    if dfns_info.suffix is not None:
        if is_defined:
            assert does_contain is True
        else:
            assert does_contain is False
    else:
        assert does_contain is None


@pytest.mark.parametrize(
    ["attr_name", "is_defined"],
    [
        ("defined", True),
        ("_defined", True),
        ("notdefined", False),
        ("_notdefined", False),
    ],
)
def test_hasattr(
    dfns: Type[DefinedNamespace], attr_name: str, is_defined: bool
) -> None:
    dfns_info = get_dfns_info(dfns)
    if dfns_info.suffix is not None:
        logging.debug("dfns_info = %s", dfns_info)
    if dfns_info.has_attrs is False:
        is_defined = False
    has_attr: Optional[bool] = None
    has_attr = hasattr(dfns, attr_name)
    if dfns_info.suffix is not None and (is_defined or dfns._fail is False):
        assert has_attr is True
    else:
        assert has_attr is False


def test_dir(dfns: Type[DefinedNamespace]) -> None:
    dfns_info = get_dfns_info(dfns)
    does_contain: Optional[bool] = None
    with ExitStack() as xstack:
        # dir should work for DefinedNamespace as this is called by sphinx to
        # document it.
        if dfns_info.suffix is None and dfns is not DefinedNamespace:
            xstack.enter_context(pytest.raises(AttributeError))
        attrs = list(dir(dfns))
    if dfns_info.suffix is not None:
        if dfns_info.has_attrs:
            assert set(attrs) == {
                URIRef(f"{prefix}{dfns_info.suffix}defined"),
                URIRef(f"{prefix}{dfns_info.suffix}_defined"),
            }
        else:
            assert list(attrs) == []
    else:
        assert does_contain is None



from rdflib import (  # noqa: I001
    BNode,
    ConjunctiveGraph,
    Dataset,
    Graph,
    Literal,
    Namespace,
    URIRef,
)
from rdflib.extras.infixowl import Class, Property, classOrTerm
from rdflib.extras.infixowl import generateQName as generate_qname  # noqa: N813
from rdflib.extras.infixowl import propertyOrIdentifier

EXNS = Namespace("http://example.org/vocab/")
PZNS = Namespace(
    "http://www.co-ode.org/ontologies/pizza/2005/10/18/classified/pizza.owl#"
)

mansyn = """
    Prefix: : <http://ex.com/owl/families#>
    Prefix: g: <http://ex.com/owl2/families#>

    Ontology: <http://example.com/owl/families> <http://example.com/owl/families-v1>
      Import: <http://ex.com/owl2/families.owl>
      Annotations: creator John,
                   Annotations: rdfs:comment "Creation Year"
                     creationYear 2008,
                   mainClass Person

      ObjectProperty: hasWife
        Annotations: ...
        Characteristics: Functional, InverseFunctional, Reflexive, Irreflexive, Asymmetric, Transitive
        Domain: Annotations: rdfs:comment "General domain",
                             creator John
                  Person,
                Annotations: rdfs:comment "More specific domain"
                  Man
        Range: Person, Woman
        SubPropertyOf: hasSpouse, loves
        EquivalentTo: isMarriedTo ,...
        DisjointWith: hates ,...
        InverseOf: hasSpouse, inverse hasSpouse
        SubPropertyChain: Annotations: ... hasChild o hasParent o...

      DataProperty: hasAge
        Annotations: ...
        Characteristics: Functional
        Domain: Person ,...
        Range: integer ,...
        SubPropertyOf: hasVerifiedAge ,...
        EquivalentTo: hasAgeInYears ,...
        DisjointWith: hasSSN ,...

      AnnotationProperty: creator
        Annotations: ...
        Domain: Person ,...
        Range: integer ,...
        SubPropertyOf: initialCreator ,...

      Datatype: NegInt
        Annotations: ...
        EquivalentTo: integer[< 0]

      Class: Person
        Annotations: ...
        SubClassOf: owl:Thing that hasFirstName exactly 1 and hasFirstName only string[minLength 1]  ,...
        SubClassOf: hasAge exactly 1 and hasAge only not NegInt,...
        SubClassOf: hasGender exactly 1 and hasGender only {female , male} ,...
        SubClassOf: hasSSN max 1, hasSSN min 1
        SubClassOf: not hates Self, ...
        EquivalentTo: g:People ,...
        DisjointWith: g:Rock , g:Mineral ,...
        DisjointUnionOf: Annotations: ... Child, Adult
        HasKey: Annotations: ... hasSSN

      Individual: John
        Annotations: ...
        Types: Person , hasFirstName value "John" or hasFirstName value "Jack"^^xsd:string
        Facts: hasWife Mary, not hasChild Susan, hasAge 33, hasChild _:child1
        SameAs: Jack ,...
        DifferentFrom: Susan ,...

      Individual: _:child1
        Annotations: ...
        Types: Person ,...
        Facts: hasChild Susan ,...

      DisjointClasses: Annotations: ... g:Rock, g:Scissor, g:Paper
      EquivalentProperties: Annotations: ... hates, loathes, despises
      DisjointProperties: Annotations: ... hates, loves, indifferent
      EquivalentProperties: Annotations: ... favoriteNumber, g:favouriteNumber, g:favouriteInteger
      DisjointProperties: Annotations: ... favoriteInteger, favouriteReal
      SameIndividual: Annotations: ... John, Jack, Joe, Jim
      DifferentIndividuals: Annotations: ... John, Susan, Mary, Jill
"""


mansynt_class = """
    /**
    * @rdfs:comment A vegetarian pizza is a pizza that only has cheese toppings
    * and tomato toppings.
    *
    * @rdfs:label Pizza [en]
    * @rdfs:label Pizza [pt]
    */
    Class: VegetarianPizza
    EquivalentTo:
    Pizza and
    not (hasTopping some FishTopping) and
    not (hasTopping some MeatTopping)
    DisjointWith:
    NonVegetarianPizza
"""


def test_generateqname_using_conjunctivegraph() -> None:
    g = ConjunctiveGraph()
    g.bind("ex", EXNS)

    assert (
        generate_qname(g, URIRef("http://example.org/vocab/ontology")) == "ex:ontology"
    )


def test_generateqname_using_dataset() -> None:
    g = Dataset()
    g.bind("ex", EXNS)

    assert (
        generate_qname(g, URIRef("http://example.org/vocab/ontology")) == "ex:ontology"
    )


def test_generateqname_using_graph() -> None:
    g = Graph()
    g.bind("ex", EXNS)

    assert (
        generate_qname(g, URIRef("http://example.org/vocab/ontology")) == "ex:ontology"
    )


def test_generateqname_using_graph_and_uriref() -> None:
    g = Graph()
    g.bind("ex", EXNS)

    res = generate_qname(g, URIRef("http://example.org/vocab/ontology"))

    assert res == "ex:ontology"


def test_generateqname_using_graph_and_bnode() -> None:
    g = Graph()
    g.bind("ex", EXNS)

    res = generate_qname(g, BNode("http://example.org/vocab/ontology"))

    assert res == "ex:ontology"

    res = generate_qname(g, BNode("urn:example:foo"))

    assert res == "ns1:foo"


def test_generateqname_using_graph_and_property() -> None:
    g = Graph()
    g.bind("ex", EXNS)

    iswoman_property = Property(identifier=EXNS.iswoman, graph=g)

    res = generate_qname(g, iswoman_property)

    assert res == "ex:iswoman"


def test_class_or_term() -> None:
    uriref_term = URIRef("http://example.org/vocab/D")
    bnode_term = BNode("http://example.org/vocab/D")
    literal_term = Literal("http://example.org/vocab/D")

    assert classOrTerm(Class(EXNS.D)) == uriref_term
    assert classOrTerm(uriref_term) == uriref_term
    assert classOrTerm(bnode_term) == bnode_term
    assert classOrTerm(literal_term) == literal_term


def test_property_or_identifier() -> None:
    g = Graph()
    g.bind("ex", EXNS)

    is_a_woman = Property(identifier=EXNS.iswoman, graph=g)

    assert propertyOrIdentifier(is_a_woman) == URIRef(
        "http://example.org/vocab/iswoman"
    )

    uriref_term = URIRef("http://example.org/vocab/D")

    assert propertyOrIdentifier(uriref_term) == URIRef("http://example.org/vocab/D")



from rdflib import Graph
from rdflib.term import BNode, URIRef, rdflib_skolem_genid


def test():
    """Test skolemised URI query retrieves expected results.

    Issue: https://github.com/RDFLib/rdflib/issues/1808
    """

    g = Graph()
    g.parse(data='[] <urn:prop> "val" .', format="turtle")
    for s, p, o in g:
        assert isinstance(s, BNode)

    gs = g.skolemize()
    for s, p, o in gs:
        assert isinstance(s, URIRef) and s.__contains__(rdflib_skolem_genid)

    query_with_iri = "select ?p ?o {{ <{}> ?p ?o }}".format(s)
    query_for_all = "select ?s ?p ?o { ?s ?p ?o }"

    count = 0
    for row in gs.query(query_with_iri):
        count += 1
    assert count == 1

    count = 0
    for row in gs.query(query_for_all):
        count += 1
    assert count == 1

    gp = Graph()
    gp.parse(data=gs.serialize(format="turtle"), format="turtle")

    count = 0
    for row in gp.query(query_with_iri):
        count += 1
    assert count == 1

    count = 0
    for row in gp.query(query_for_all):
        count += 1
    assert count == 1



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Collection, Set, Tuple, Type, Union, cast

import pytest
from _pytest.mark.structures import ParameterSet

import rdflib
from rdflib import Graph
from rdflib.compare import graph_diff
from rdflib.graph import ConjunctiveGraph, Dataset
from rdflib.namespace import FOAF, RDF
from rdflib.term import BNode, Literal
from test.utils import (
    COLLAPSED_BNODE,
    BNodeHandling,
    GHQuad,
    GHTriple,
    GraphHelper,
    MarksType,
    MarkType,
)

if TYPE_CHECKING:
    from rdflib.graph import _TripleType

"""Test for graph_diff - much more extensive testing
would certainly be possible"""

_TripleSetType = Set["_TripleType"]


class TestDiff:
    """Unicode literals for graph_diff test
    (issue 151)"""

    def test_a(self):
        """with bnode"""
        g = rdflib.Graph()
        g.add((rdflib.BNode(), rdflib.URIRef("urn:p"), rdflib.Literal("\xe9")))

        graph_diff(g, g)

    def test_b(self):
        """Curiously, this one passes, even before the fix in issue 151"""

        g = rdflib.Graph()
        g.add((rdflib.URIRef("urn:a"), rdflib.URIRef("urn:p"), rdflib.Literal("\xe9")))

        graph_diff(g, g)

    @pytest.mark.xfail()
    def test_subsets(self) -> None:
        """
        This test verifies that `graph_diff` returns the correct values
        for two graphs, `g0` and `g1` where the triples in `g0` is a
        subset of the triples in `g1`.

        The expectation is that graph_diff reports that there are no
        triples only in `g0`, and that there are triples that occur in both
        `g0` and `g1`, and that there are triples only in `g1`.
        """
        g0_ts: _TripleSetType = set()
        bnode = BNode()
        g0_ts.update(
            {
                (bnode, FOAF.name, Literal("Golan Trevize")),
                (bnode, RDF.type, FOAF.Person),
            }
        )
        g0 = Graph()
        g0 += g0_ts

        g1_ts: _TripleSetType = set()
        bnode = BNode()
        g1_ts.update(
            {
                *g0_ts,
                (bnode, FOAF.name, Literal("Janov Pelorat")),
                (bnode, RDF.type, FOAF.Person),
            }
        )
        g1 = Graph()
        g1 += g1_ts

        result = graph_diff(g0, g1)
        in_both, in_first, in_second = GraphHelper.triple_sets(result)
        assert in_first == set()
        assert len(in_second) > 0
        assert len(in_both) > 0


_ElementSetType = Union[Collection[GHTriple], Collection[GHQuad]]

_ElementSetTypeOrStr = Union[_ElementSetType, str]


@dataclass
class GraphDiffCase:
    graph_type: Type[Graph]
    format: str
    lhs: str
    rhs: str
    expected_result: Tuple[
        _ElementSetTypeOrStr, _ElementSetTypeOrStr, _ElementSetTypeOrStr
    ]
    marks: MarkType = field(default_factory=lambda: cast(MarksType, list()))

    def as_element_set(self, value: _ElementSetTypeOrStr) -> _ElementSetType:
        if isinstance(value, str):
            graph = self.graph_type()
            graph.parse(data=value, format=self.format)
            if isinstance(graph, ConjunctiveGraph):
                return GraphHelper.quad_set(graph, BNodeHandling.COLLAPSE)
            else:
                return GraphHelper.triple_set(graph, BNodeHandling.COLLAPSE)
        return value

    def expected_in_both_set(self) -> _ElementSetType:
        return self.as_element_set(self.expected_result[0])

    def expected_in_lhs_set(self) -> _ElementSetType:
        return self.as_element_set(self.expected_result[1])

    def expected_in_rhs_set(self) -> _ElementSetType:
        return self.as_element_set(self.expected_result[2])

    def as_params(self) -> ParameterSet:
        return pytest.param(self, marks=self.marks)


@pytest.mark.parametrize(
    "test_case",
    [
        GraphDiffCase(
            Graph,
            format="turtle",
            lhs="""
            @prefix eg: <example:> .
            _:a _:b _:c .
            eg:o0 eg:p0 eg:s0 .
            eg:o1 eg:p1 eg:s1 .
            """,
            rhs="""
            @prefix eg: <example:> .
            eg:o0 eg:p0 eg:s0 .
            eg:o1 eg:p1 eg:s1 .
            """,
            expected_result=(
                """
            @prefix eg: <example:> .
            eg:o0 eg:p0 eg:s0 .
            eg:o1 eg:p1 eg:s1 .
            """,
                {(COLLAPSED_BNODE, COLLAPSED_BNODE, COLLAPSED_BNODE)},
                "",
            ),
        ),
        GraphDiffCase(
            Graph,
            format="turtle",
            lhs="""
            @prefix eg: <example:> .
            eg:o0 eg:p0 eg:s0 .
            eg:o1 eg:p1 eg:s1 .
            """,
            rhs="""
            @prefix eg: <example:> .
            eg:o0 eg:p0 eg:s0 .
            eg:o1 eg:p1 eg:s1 .
            """,
            expected_result=(
                """
            @prefix eg: <example:> .
            eg:o0 eg:p0 eg:s0 .
            eg:o1 eg:p1 eg:s1 .
            """,
                "",
                "",
            ),
        ),
        GraphDiffCase(
            Dataset,
            format="trig",
            lhs="""
            @prefix eg: <example:> .
            eg:o0 eg:p0 eg:s0 .
            eg:o1 eg:p1 eg:s1 .
            """,
            rhs="""
            @prefix eg: <example:> .
            eg:o0 eg:p0 eg:s0 .
            eg:o1 eg:p1 eg:s1 .
            """,
            expected_result=(
                """
            @prefix eg: <example:> .
            eg:o0 eg:p0 eg:s0 .
            eg:o1 eg:p1 eg:s1 .
            """,
                "",
                "",
            ),
            marks=pytest.mark.xfail(
                reason="quads are not supported", raises=ValueError
            ),
        ).as_params(),
    ],
)
def test_assert_sets_equal(test_case: GraphDiffCase):
    """
    GraphHelper.sets_equals and related functions work correctly in both
    positive and negative cases.
    """
    lhs_graph: Graph = test_case.graph_type()
    lhs_graph.parse(data=test_case.lhs, format=test_case.format)

    rhs_graph: Graph = test_case.graph_type()
    rhs_graph.parse(data=test_case.rhs, format=test_case.format)

    in_both, in_lhs, in_rhs = graph_diff(lhs_graph, rhs_graph)
    in_both_set = GraphHelper.triple_or_quad_set(in_both, BNodeHandling.COLLAPSE)
    in_lhs_set = GraphHelper.triple_or_quad_set(in_lhs, BNodeHandling.COLLAPSE)
    in_rhs_set = GraphHelper.triple_or_quad_set(in_rhs, BNodeHandling.COLLAPSE)

    assert test_case.expected_in_both_set() == in_both_set
    assert test_case.expected_in_lhs_set() == in_lhs_set
    assert test_case.expected_in_rhs_set() == in_rhs_set

    # Diff should be symetric
    in_rboth, in_rlhs, in_rrhs = graph_diff(rhs_graph, lhs_graph)
    in_rboth_set = GraphHelper.triple_or_quad_set(in_rboth, BNodeHandling.COLLAPSE)
    in_rlhs_set = GraphHelper.triple_or_quad_set(in_rlhs, BNodeHandling.COLLAPSE)
    in_rrhs_set = GraphHelper.triple_or_quad_set(in_rrhs, BNodeHandling.COLLAPSE)

    assert test_case.expected_in_both_set() == in_rboth_set
    assert test_case.expected_in_rhs_set() == in_rlhs_set
    assert test_case.expected_in_lhs_set() == in_rrhs_set



from rdflib.plugins.shared.jsonld.util import norm_url


def test_norm_urn():
    assert norm_url("urn:ns:test", "/one") == "urn:ns:test/one"
    assert norm_url("urn:ns:test/path/", "two") == "urn:ns:test/path/two"
    assert norm_url("urn:ns:test/path", "two") == "urn:ns:test/two"
    assert norm_url("urn:ns:test", "three") == "urn:ns:test/three"
    assert norm_url("urn:ns:test/path#", "four") == "urn:ns:test/four"
    assert norm_url("urn:ns:test/path1/path2/", "../path3") == "urn:ns:test/path1/path3"
    assert norm_url("urn:ns:test/path1/path2/", "/path3") == "urn:ns:test/path3"
    assert (
        norm_url("urn:ns:test/path1/path2/", "http://example.com")
        == "http://example.com"
    )
    assert (
        norm_url("urn:ns:test/path1/path2/", "urn:another:test/path")
        == "urn:another:test/path"
    )
    assert norm_url("urn:ns:test/path", "#four") == "urn:ns:test/path#four"
    assert norm_url("urn:ns:test/path/", "#four") == "urn:ns:test/path/#four"



from __future__ import annotations

import logging
import os
from contextlib import ExitStack
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple, Union

import pytest

from rdflib import Graph
from rdflib.tools.chunk_serializer import serialize_in_chunks
from test.data import TEST_DATA_DIR
from test.utils import GraphHelper
from test.utils.graph import cached_graph
from test.utils.namespace import MF
from test.utils.path import ctx_chdir

if TYPE_CHECKING:
    from builtins import ellipsis

logger = logging.getLogger(__name__)


def test_chunk_by_triples(tmp_path: Path):
    g = cached_graph((TEST_DATA_DIR / "suites/w3c/trig/manifest.ttl",))
    # this graph has 2,848 triples

    # serialize into chunks file with 100 triples each
    serialize_in_chunks(
        g, max_triples=100, file_name_stem="chunk_100", output_dir=tmp_path
    )

    # count the resultant .nt files, should be math.ceil(2848 / 100) = 25
    assert len(list(tmp_path.glob("*.nt"))) == 25

    # check, when a graph is made from the chunk files, it's isomorphic with original
    g2 = Graph()
    for f in tmp_path.glob("*.nt"):
        g2.parse(f, format="nt")

    assert g.isomorphic(g2), "Reconstructed graph is not isomorphic with original"


def test_chunk_by_size(tmp_path: Path):
    g = cached_graph((TEST_DATA_DIR / "suites/w3c/trig/manifest.ttl",))
    # as an NT file, this graph is 323kb

    # serialize into chunks file of > 50kb each
    serialize_in_chunks(
        g, max_file_size_kb=50, file_name_stem="chunk_50k", output_dir=tmp_path
    )

    # check all files are size < 50kb, with a margin up to 60kb
    for f in Path(tmp_path).glob("*.nt"):
        assert os.path.getsize(f) < 50000

    # check, when a graph is made from the chunk files, it's isomorphic with original
    g2 = Graph()
    for f in Path(tmp_path).glob("*.nt"):
        g2.parse(f, format="nt")

    assert g.isomorphic(g2), "Reconstructed graph is not isomorphic with original"


@pytest.mark.parametrize(
    [
        "test_graph_path",
        "max_triples",
        "max_file_size_kb",
        "write_prefixes",
        "set_output_dir",
        "expected_file_count",
    ],
    [
        (TEST_DATA_DIR / "defined_namespaces/mf.ttl", ..., ..., False, True, 1),
        (TEST_DATA_DIR / "defined_namespaces/mf.ttl", ..., ..., True, False, 2),
        (TEST_DATA_DIR / "defined_namespaces/mf.ttl", ..., 5, True, False, (3, 7)),
        (TEST_DATA_DIR / "defined_namespaces/mf.ttl", ..., 1, False, True, (15, 25)),
        (TEST_DATA_DIR / "defined_namespaces/mf.ttl", 10000, 1, False, True, (15, 25)),
        (TEST_DATA_DIR / "defined_namespaces/mf.ttl", 20, ..., False, True, 5),
        (TEST_DATA_DIR / "defined_namespaces/mf.ttl", 20, ..., True, True, 6),
        (TEST_DATA_DIR / "defined_namespaces/mf.ttl", 100, ..., True, True, 2),
        (TEST_DATA_DIR / "defined_namespaces/mf.ttl", 100, ..., False, True, 1),
    ],
)
def test_chuking(
    tmp_path: Path,
    test_graph_path: Path,
    max_triples: Union[ellipsis, int],
    max_file_size_kb: Union[ellipsis, int, None],
    write_prefixes: bool,
    set_output_dir: bool,
    expected_file_count: Optional[Union[int, Tuple[Optional[int], Optional[int]]]],
) -> None:
    test_graph = cached_graph((test_graph_path,))
    kwargs: Dict[str, Any] = {"write_prefixes": write_prefixes}
    if max_triples is not ...:
        kwargs["max_triples"] = max_triples
    if max_file_size_kb is not ...:
        kwargs["max_file_size_kb"] = max_file_size_kb
    logger.debug("kwargs = %s", kwargs)
    with ExitStack() as xstack:
        if set_output_dir:
            kwargs["output_dir"] = tmp_path
        else:
            xstack.enter_context(ctx_chdir(tmp_path))
        serialize_in_chunks(test_graph, **kwargs)

    # set the values to defaults if they were elided in test parameters.
    if max_file_size_kb is ...:
        max_file_size_kb = None
    if max_triples is ...:
        max_triples = 10000

    stem = "chunk"

    output_paths = set(item.relative_to(tmp_path) for item in tmp_path.glob("**/*"))
    # output_filenames = set(f"{item}" for item in output_paths)

    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("tmp_path = %s files = %s", tmp_path, output_paths)

    if isinstance(expected_file_count, tuple):
        if expected_file_count[0] is not None:
            assert expected_file_count[0] <= len(output_paths)
        if expected_file_count[1] is not None:
            assert expected_file_count[1] >= len(output_paths)
    elif isinstance(expected_file_count, int):
        assert expected_file_count == len(output_paths)

    recovered_graph = Graph(bind_namespaces="none")

    if write_prefixes is True:
        prefixes_path = Path(f"{stem}_000000.ttl")
        assert prefixes_path in output_paths
        output_paths.remove(prefixes_path)
        recovered_graph.parse(tmp_path / prefixes_path, format="turtle")
        namespaces_data = (tmp_path / prefixes_path).read_text("utf-8")
        assert f"{MF}" in namespaces_data

    if len(output_paths) == 1:
        all_file = Path(f"{stem}_all.nt")
        assert all_file in output_paths
        all_file = tmp_path / all_file
        file_bytes = all_file.read_bytes()
        recovered_graph.parse(all_file, format="nt")
        if isinstance(max_file_size_kb, int):
            assert len(file_bytes) <= (max_file_size_kb * 1000)
        elif isinstance(max_triples, int):
            assert len(recovered_graph) <= max_triples

    elif max_file_size_kb is not None:
        assert isinstance(max_file_size_kb, int)
        for output_path in output_paths:
            output_path = tmp_path / output_path
            file_bytes = output_path.read_bytes()
            assert len(file_bytes) <= (max_file_size_kb * 1000)
            logger.debug("reading %s", output_path)
            recovered_graph.parse(output_path, format="nt")
    else:
        assert isinstance(max_triples, int)
        for output_path in output_paths:
            output_path = tmp_path / output_path
            file_bytes = output_path.read_bytes()
            triple_count = len(recovered_graph)
            logger.debug("reading %s", output_path)
            recovered_graph.parse(output_path, format="nt")
            extra_triples = len(recovered_graph) - triple_count
            assert extra_triples <= max_triples

    logger.debug("checking isomorphism")
    GraphHelper.assert_isomorphic(test_graph, recovered_graph)



from rdflib.graph import ConjunctiveGraph


def test_turtle_namespace_prefixes():
    g = ConjunctiveGraph()
    n3 = """
    @prefix _9: <http://data.linkedmdb.org/resource/movie/> .
    @prefix p_9: <urn:test:> .
    @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .

    p_9:a p_9:b p_9:c .

    <http://data.linkedmdb.org/resource/director/1> a
    <http://data.linkedmdb.org/resource/movie/director>;
        rdfs:label "Cecil B. DeMille (Director)";
        _9:director_name "Cecil B. DeMille" ."""

    g.parse(data=n3, format="n3")
    turtle = g.serialize(format="turtle")

    # Check round-tripping, just for kicks.
    g = ConjunctiveGraph()
    g.parse(data=turtle, format="turtle")
    # Shouldn't have got to here
    s = g.serialize(format="turtle", encoding="latin-1")

    assert b"@prefix _9" not in s



from rdflib import events


class AddedEvent(events.Event):
    pass


class RemovedEvent(events.Event):
    pass


def subscribe_to(source, target):
    target.subscribe(AddedEvent, source._add_handler)
    target.subscribe(RemovedEvent, source._remove_handler)


def subscribe_all(caches):
    for cache in caches:
        for other in caches:
            if other != cache:
                subscribe_to(cache, other)


class Cache(events.Dispatcher):
    def __init__(self, data=None):
        if data is None:
            data = {}
        self._data = data
        self.subscribe(AddedEvent, self._add_handler)
        self.subscribe(RemovedEvent, self._remove_handler)

    def _add_handler(self, event):
        self._data[event.key] = event.value

    def _remove_handler(self, event):
        del self._data[event.key]

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self.dispatch(AddedEvent(key=key, value=value))

    def __delitem__(self, key):
        self.dispatch(RemovedEvent(key=key))

    def __contains__(self, key):
        return key in self._data

    has_key = __contains__


class TestEvent:
    def testEvents(self):  # noqa: N802
        c1 = Cache()
        c2 = Cache()
        c3 = Cache()
        subscribe_all([c1, c2, c3])
        c1["bob"] = "uncle"
        assert c2["bob"] == "uncle"
        assert c3["bob"] == "uncle"
        del c3["bob"]
        assert ("bob" in c1) is False
        assert ("bob" in c2) is False



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from typing import (
    Callable,
    Collection,
    Generator,
    Iterable,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from urllib.parse import urljoin

import pytest
from _pytest.mark.structures import Mark, MarkDecorator, ParameterSet

from rdflib.graph import Graph
from rdflib.namespace import RDF
from rdflib.term import IdentifiedNode, Identifier, URIRef
from test.utils import MarkListType, marks_to_list
from test.utils.graph import GraphSource, GraphSourceType
from test.utils.iri import URIMapper
from test.utils.namespace import MF

POFilterType = Tuple[Optional[URIRef], Optional[URIRef]]
POFiltersType = Iterable[POFilterType]

MarkType = Union[MarkDecorator, Collection[Union[MarkDecorator, Mark]]]
MarksDictType = Mapping[
    str, Union[MarkDecorator, Collection[Union[MarkDecorator, Mark]]]
]
ManifestEntryMarkerType = Callable[["ManifestEntry"], Optional[MarkType]]
IdentifierT = TypeVar("IdentifierT", bound=Identifier)


@dataclass
class ManifestEntry:
    manifest: Manifest
    identifier: URIRef
    type: IdentifiedNode = field(init=False)
    action: Optional[IdentifiedNode] = field(init=False)
    result: Optional[IdentifiedNode] = field(init=False)
    result_cardinality: Optional[URIRef] = field(init=False)

    def __post_init__(self) -> None:
        type = self.value(RDF.type, IdentifiedNode)
        assert type is not None
        self.type = type

        self.action = self.value(MF.action, IdentifiedNode)
        self.result = self.value(MF.result, IdentifiedNode)
        self.result_cardinality = self.value(MF.resultCardinality, URIRef)
        if self.result_cardinality is not None:
            assert self.result_cardinality == MF.LaxCardinality

    @property
    def graph(self) -> Graph:
        return self.manifest.graph

    @property
    def uri_mapper(self) -> URIMapper:
        return self.manifest.uri_mapper

    def param(
        self,
        mark_dict: Optional[MarksDictType] = None,
        markers: Optional[Iterable[ManifestEntryMarkerType]] = None,
    ) -> ParameterSet:
        id = f"{self.identifier}"
        marks: MarkListType = []
        if mark_dict is not None:
            marks = marks_to_list(mark_dict.get(id, marks))
        if markers is not None:
            for marker in markers:
                opt_marks = marker(self)
                if opt_marks is not None:
                    opt_marks = marks_to_list(opt_marks)
                    marks.extend(opt_marks)
        return pytest.param(self, id=f"{self.identifier}", marks=marks)

    def value(
        self, predicate: Identifier, value_type: Type[IdentifierT]
    ) -> Optional[IdentifierT]:
        value = self.graph.value(self.identifier, predicate)
        if value is not None:
            assert isinstance(value, value_type)
        return value

    def check_filters(self, filters: POFiltersType) -> bool:
        for filter in filters:
            if (self.identifier, filter[0], filter[1]) in self.graph:
                return True
        return False


@dataclass
class Manifest:
    uri_mapper: URIMapper
    graph: Graph
    identifier: IdentifiedNode
    report_prefix: Optional[str] = None

    @classmethod
    def from_graph(
        cls,
        uri_mapper: URIMapper,
        graph: Graph,
        report_prefix: Optional[str] = None,
    ) -> Generator[Manifest, None, None]:
        for identifier in graph.subjects(RDF.type, MF.Manifest):
            assert isinstance(identifier, IdentifiedNode)
            manifest = Manifest(
                uri_mapper,
                graph,
                identifier,
                report_prefix,
            )
            yield manifest
            yield from manifest.included()

    @classmethod
    def from_sources(
        cls,
        uri_mapper: URIMapper,
        *sources: GraphSourceType,
        report_prefix: Optional[str] = None,
    ) -> Generator[Manifest, None, None]:
        for source in sources:
            logging.debug("source(%s) = %r", id(source), source)
            source = GraphSource.from_source(source)
            source_path_uri = source.path.absolute().as_uri()
            local_base = urljoin(source_path_uri, ".")
            public_id = uri_mapper.to_remote(local_base)
            logging.debug(
                "source = %s, source_path_uri = %s, local_base = %s, public_id = %s",
                source,
                source_path_uri,
                local_base,
                public_id,
            )
            graph: Graph = source.load(public_id=public_id)
            yield from cls.from_graph(
                uri_mapper,
                graph,
                report_prefix,
            )

    def included(self) -> Generator[Manifest, None, None]:
        for includes in self.graph.objects(self.identifier, MF.include):
            for include in self.graph.items(includes):
                assert isinstance(include, str)
                include_local_path = self.uri_mapper.to_local_path(include)
                yield from Manifest.from_sources(
                    self.uri_mapper,
                    include_local_path,
                    report_prefix=self.report_prefix,
                )

    def entires(
        self,
        entry_type: Type[ManifestEntryT],
        exclude: Optional[POFiltersType] = None,
        include: Optional[POFiltersType] = None,
    ) -> Generator[ManifestEntryT, None, None]:
        for entries in self.graph.objects(self.identifier, MF.entries):
            for entry_iri in self.graph.items(entries):
                assert isinstance(entry_iri, URIRef)
                entry = entry_type(self, entry_iri)
                if exclude is not None and entry.check_filters(exclude):
                    continue
                if include is not None and not entry.check_filters(include):
                    continue
                yield entry

    def params(
        self,
        entry_type: Type[ManifestEntryT],
        exclude: Optional[POFiltersType] = None,
        include: Optional[POFiltersType] = None,
        mark_dict: Optional[MarksDictType] = None,
        markers: Optional[Iterable[ManifestEntryMarkerType]] = None,
    ) -> Generator[ParameterSet, None, None]:
        for entry in self.entires(entry_type, exclude, include):
            yield entry.param(mark_dict, markers)


def params_from_sources(
    uri_mapper: URIMapper,
    entry_type: Type[ManifestEntryT],
    *sources: GraphSourceType,
    exclude: Optional[POFiltersType] = None,
    include: Optional[POFiltersType] = None,
    mark_dict: Optional[MarksDictType] = None,
    markers: Optional[Iterable[ManifestEntryMarkerType]] = None,
    report_prefix: Optional[str] = None,
) -> Generator[ParameterSet, None, None]:
    for manifest in Manifest.from_sources(
        uri_mapper, *sources, report_prefix=report_prefix
    ):
        yield from manifest.params(entry_type, include, exclude, mark_dict, markers)


ManifestEntryT = TypeVar("ManifestEntryT", bound=ManifestEntry)



import os

import pytest

import rdflib


@pytest.mark.skipif(
    condition=(os.name == "nt"),
    reason="No os.fork() and/or os.pipe() on this platform, skipping",
)
def test_bnode_id_differs_in_fork():
    """Checks that os.fork()ed child processes produce a
    different sequence of BNode ids from the parent process.
    """
    r, w = os.pipe()  # these are file descriptors, not file objects
    pid = os.fork()
    if pid:
        pb1 = rdflib.term.BNode()
        os.close(w)  # use os.close() to close a file descriptor
        r = os.fdopen(r)  # turn r into a file object
        txt = r.read()
        os.waitpid(pid, 0)  # make sure the child process gets cleaned up
        r.close()
    else:
        os.close(r)
        w = os.fdopen(w, "w")
        cb = rdflib.term.BNode()
        w.write(cb)
        w.close()
        os._exit(0)
    assert txt != str(
        pb1
    ), "Parent process BNode id: " + "%s, child process BNode id: %s" % (
        txt,
        str(pb1),
    )



# https://github.com/RDFLib/rdflib-jsonld/blob/feature/json-ld-1.1/rdflib_jsonld/errors.py
class JSONLDException(ValueError):  # noqa: N818
    pass


# http://www.w3.org/TR/json-ld-api/#idl-def-JsonLdErrorCode.{code-message}
RECURSIVE_CONTEXT_INCLUSION = JSONLDException("recursive context inclusion")
INVALID_REMOTE_CONTEXT = JSONLDException("invalid remote context")
INVALID_CONTEXT_ENTRY = JSONLDException("invalid context entry")



"""
Issue 801 - Problem with prefixes created for URIs containing %20
"""

from rdflib import BNode, Graph, Literal
from test.utils.namespace import EGDO


def test_issue_801():
    g = Graph()
    g.bind("", EGDO)
    node = BNode()
    g.add((node, EGDO["first%20name"], Literal("John")))
    assert g.serialize(format="turtle").split("\n")[-3] == '[] :first%20name "John" .'



"""
Issue 920 - NTriples fails to parse URIs with only a scheme

from rdflib import Graph
g=Graph()
g.parse(data='<a:> <b:> <c:> .', format='nt') # nquads also fails

N3, by contrast, succeeds:

g.parse(data='<a:> <b:> <c:> .', format='n3')
"""

from rdflib import Graph


def test_issue_920():
    g = Graph()
    # NT tests
    g.parse(data="<a:> <b:> <c:> .", format="nt")
    g.parse(data="<http://a> <http://b> <http://c> .", format="nt")
    g.parse(data="<https://a> <http://> <http://c> .", format="nt")

    # related parser tests
    g.parse(data="<a:> <b:> <c:> .", format="turtle")
    g.parse(data="<http://a> <http://b> <http://c> .", format="turtle")
    g.parse(data="<https://a> <http://> <http://c> .", format="turtle")

    g.parse(data="<a:> <b:> <c:> .", format="n3")
    g.parse(data="<http://a> <http://b> <http://c> .", format="n3")
    g.parse(data="<https://a> <http://> <http://c> .", format="n3")



import textwrap

import pytest

from rdflib.graph import ConjunctiveGraph
from rdflib.parser import StringInputSource

prefix = textwrap.dedent(
    """\
    @prefix nie: <http://www.semanticdesktop.org/ontologies/2007/01/19/nie#> .
    @prefix nfo: <http://www.semanticdesktop.org/ontologies/2007/03/22/nfo#> .
    @prefix nco: <http://www.semanticdesktop.org/ontologies/2007/03/22/nco#> .
    @prefix nmo: <http://www.semanticdesktop.org/ontologies/2007/03/22/nmo#> .
    @prefix ncal: <http://www.semanticdesktop.org/ontologies/2007/04/02/ncal#> .
    @prefix nexif: <http://www.semanticdesktop.org/ontologies/2007/05/10/nexif#> .
    @prefix nid3: <http://www.semanticdesktop.org/ontologies/2007/05/10/nid3#> .
    @prefix dc: <http://dublincore.org/documents/2010/10/11/dces/#> .
    @prefix nmm: <http://library.gnome.org/devel/ontology/unstable/nmm-classes.html#> .
    @prefix nao: <http://www.semanticdesktop.org/ontologies/2007/08/15/nao#> .
    """
)

meta = textwrap.dedent(
    """\
a nfo:PaginatedTextDocument ;
    nie:title "SV Meldung" ;
    nco:creator [ a nco:Contact ;
    nco:fullname "nikratio"] ;
    nie:contentCreated "2011-08-10T20:12:38Z" ;
    dc:format "application/pdf" ;
    nie:description "()" ;
    nao:hasTag ?tag1 ;
    nfo:pageCount 1 ;
    nie:plainTextContent "%s" .
} } WHERE { {
?tag1 a nao:Tag ; nao:prefLabel "()" .
"""
)

test_string1 = """\
Betriebsnummer der Einzugsstelle:\nKnappschaft\n980 0000 6\nWICHTIGES DOKUMENT - SORGFÄLTIG AUFBEWAHREN!\n """


@pytest.mark.xfail(reason="Known issue with newlines in text")
def test1():
    meta1 = meta.encode("utf-8") % test_string1.encode("utf-8")
    graph = ConjunctiveGraph()
    graph.parse(
        StringInputSource(prefix + "<http://example.org/>" + meta1), format="n3"
    )


test_string2 = """\
Betriebsnummer der Einzugsstelle:
Knappschaft
980 0000 6
WICHTIGES DOKUMENT - SORGFÄLTIG AUFBEWAHREN!
"""


@pytest.mark.xfail(reason="Known issue with newlines in text")
def test2():
    meta2 = meta.encode("utf-8") % test_string2.encode("utf-8")
    graph = ConjunctiveGraph()
    graph.parse(
        StringInputSource(prefix + "<http://example.org/>" + meta2), format="n3"
    )



# rdflib documentation build configuration file, created by
# sphinx-quickstart on Fri May 15 15:03:54 2009.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.
# https://www.sphinx-doc.org/en/master/usage/configuration.html
from __future__ import annotations

import logging
import os
import re
import sys
from typing import Any, Dict

import sphinx
import sphinx.application

import rdflib

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.append(os.path.abspath(".."))
sys.path.append(os.path.abspath(".."))

# -- General configuration -----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
# extensions = ['sphinx.ext.autodoc', 'sphinx.ext.todo', 'sphinx.ext.doctest']
extensions = [
    "sphinxcontrib.apidoc",
    "sphinx.ext.autodoc",
    #'sphinx.ext.autosummary',
    "sphinx_autodoc_typehints",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    "myst_parser",
    "sphinx.ext.autosectionlabel",
]

# https://github.com/sphinx-contrib/apidoc/blob/master/README.rst#configuration
apidoc_module_dir = "../rdflib"
apidoc_output_dir = "apidocs"

# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html
autodoc_default_options = {"special-members": True}
autodoc_inherit_docstrings = True

# https://github.com/tox-dev/sphinx-autodoc-typehints
always_document_param_types = True

autosummary_generate = True

autosectionlabel_prefix_document = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# epydoc_mapping = {
#   '/_static/api/': [r'rdflib\.'],
#   }

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
source_encoding = "utf-8"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "rdflib"
copyright = "2009 - 2024, RDFLib Team"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.


# Find version. We have to do this because we can't import it in Python 3 until
# its been automatically converted in the setup process.
# UPDATE: This function is no longer used; once builds are confirmed to succeed, it
#         can/should be removed. --JCL 2022-12-30
def find_version(filename):
    _version_re = re.compile(r'__version__ = "(.*)"')
    for line in open(filename):
        version_match = _version_re.match(line)
        if version_match:
            return version_match.group(1)


# The full version, including alpha/beta/rc tags.
release = rdflib.__version__
# The short X.Y version.
version = re.sub("[0-9]+\\.[0-9]\\..*", "\1", release)

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of documents that shouldn't be included in the build.
# unused_docs = []

# List of directories, relative to source directory, that shouldn't be searched
# for source files.
exclude_trees = ["_build", "draft"]

# The reST default role (used for this markup: `text`) to use for all documents.
default_role = "py:obj"

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
html_theme = "armstrong"


# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = [
    "_themes",
]

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None
html_logo = "_static/RDFlib.png"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/RDFlib.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_use_modindex = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# If nonempty, this is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = ''

# Output file base name for HTML help builder.
htmlhelp_basename = "rdflibdoc"


# -- Options for LaTeX output --------------------------------------------------

# The paper size ('letter' or 'a4').
# latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
# latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
# latex_documents = [
#     ("index", "rdflib.tex", "rdflib Documentation", "RDFLib Team", "manual"),
# ]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# Additional stuff for the LaTeX preamble.
# latex_preamble = ''

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_use_modindex = True


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3.8", None),
}

html_experimental_html5_writer = True

needs_sphinx = "4.1.2"

suppress_warnings = [
    # This is here to prevent:
    #  "WARNING: more than one target found for cross-reference"
    "ref.python",
    "autosectionlabel.*",
]

sphinx_version = tuple(int(part) for part in sphinx.__version__.split("."))


nitpicky = True

nitpick_ignore = [
    ("py:class", "urllib.response.addinfourl"),
    ("py:class", "importlib.metadata.EntryPoint"),
    ("py:class", "xml.dom.minidom.Document"),
    ("py:class", "xml.dom.minidom.DocumentFragment"),
    ("py:class", "isodate.duration.Duration"),
    ("py:class", "pyparsing.core.TokenConverter"),
    ("py:class", "pyparsing.results.ParseResults"),
    ("py:class", "pyparsing.core.ParserElement"),
]


def autodoc_skip_member_handler(
    app: sphinx.application.Sphinx,
    what: str,
    name: str,
    obj: Any,
    skip: bool,
    options: Dict[str, Any],
):
    """
    This function will be called by Sphinx when it is deciding whether to skip a
    member of a class or module.
    """
    # https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#event-autodoc-skip-member
    if (
        app.env.docname == "apidocs/rdflib"
        and what == "module"
        and type(obj).__name__.endswith("DefinedNamespaceMeta")
    ):
        # Don't document namespaces in the `rdflib` module, they will be
        # documented in the `rdflib.namespace` module instead and Sphinx does
        # not like when these are documented in two places.
        #
        # An example of the WARNINGS that occur without this is:
        #
        # "WARNING: duplicate object description of rdflib.namespace._SDO.SDO,
        # other instance in apidocs/rdflib, use :noindex: for one of them"
        logging.info(
            "Skipping %s %s in %s, it will be documented in ",
            what,
            name,
            app.env.docname,
        )
        return True
    return None


# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#skipping-members
def setup(app: sphinx.application.Sphinx) -> None:
    """
    Setup the Sphinx application.
    """

    # Register a autodoc-skip-member handler so that certain members can be
    # skipped.
    app.connect("autodoc-skip-member", autodoc_skip_member_handler)



"""
This is a simple primer using some of the
example stuff in the Primer on N3:

http://www.w3.org/2000/10/swap/Primer
"""

from rdflib import ConjunctiveGraph, Literal, Namespace
from rdflib.namespace import DC, OWL

if __name__ == "__main__":
    # Firstly, it doesn't have to be so complex.
    # Here we create a "Graph" of our work.
    # Think of it as a blank piece of graph paper!

    primer = ConjunctiveGraph()
    myNS = Namespace("https://example.com/")  # noqa: N816

    primer.add((myNS.pat, myNS.knows, myNS.jo))
    # or:
    primer.add((myNS["pat"], myNS["age"], Literal(24)))

    # Now, with just that, lets see how the system
    # recorded *way* too many details about what
    # you just asserted as fact.

    from pprint import pprint

    print("All the things in the Graph:")
    pprint(list(primer))

    # just think .whatever((s, p, o))
    # here we report on what we know

    print("==================")

    print("Subjects:")
    pprint(list(primer.subjects()))
    print("Predicates:")
    pprint(list(primer.predicates()))
    print("Objects:")
    pprint(list(primer.objects()))

    print("==================")
    # and other things that make sense

    print("What we know about pat:")
    pprint(list(primer.predicate_objects(myNS.pat)))

    print("Who is what age?")
    pprint(list(primer.subject_objects(myNS.age)))

    print("==================")
    print("==================")

    # Okay, so lets now work with a bigger
    # dataset from the example, and start
    # with a fresh new graph.

    del primer
    primer = ConjunctiveGraph()

    # Lets start with a verbatim string straight from the primer text:

    mySource = """
    @prefix : <https://example.com/> .
    @prefix rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
    @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
    @prefix owl:  <http://www.w3.org/2002/07/owl#> .
    @prefix dc:  <http://purl.org/dc/elements/1.1/> .
    @prefix foo: <http://www.w3.org/2000/10/swap/Primer#>.
    @prefix swap: <http://www.w3.org/2000/10/swap/>.

    <> dc:title
      "Primer - Getting into the Semantic Web and RDF using N3".

    :pat    :knows          :jo .
    :pat    :age            24 .
    :al     :is_child_of    :pat .

    :pat    :child          :al, :chaz, :mo ;
            :age            24 ;
            :eyecolor       "blue" .


    :Person a rdfs:Class .

    :Pat a :Person .

    :Woman a rdfs:Class; rdfs:subClassOf :Person .

    :sister a rdf:Property .

    :sister rdfs:domain :Person ;
            rdfs:range :Woman .

    :Woman = foo:FemaleAdult .
    :Title a rdf:Property; = dc:title .
    """  # --- End of primer code  # noqa: N816

    # To make this go easier to spit back out...
    # technically, we already created a namespace
    # with the object init (and it added some namespaces as well)
    # By default, your main namespace is the URI of your
    # current working directory, so lets make that simpler:

    primer.bind("owl", OWL)
    primer.bind("dc", DC)
    primer.bind("swap", "http://www.w3.org/2000/10/swap/")

    # Lets load it up!

    primer.parse(data=mySource, format="n3")

    # Now you can query, either directly straight into a list:

    print()
    print("Printing bigger example's triples:")
    for i in [(x, y, z) for x, y, z in primer]:
        print(i)

    # or spit it back out (mostly) the way we created it:

    print()
    print("Printing bigger example as N3:")
    print(primer.serialize(format="n3"))

    # for more insight into things already done, lets see the namespaces

    print()
    print("Printing bigger example's namespaces:")
    for n in list(primer.namespaces()):
        print(n)

    # lets ask something about the data, using a SPARQL query

    print()
    print("Who are pat's children?")
    q = "SELECT ?child WHERE { :pat :child ?child }"
    for r in primer.query(q):
        print(r)



"""
Testing with pytest
=================

This test runner uses pytest for test discovery and running. It uses the argument
spec of pytest, but with some options pre-set. To begin with, make sure you have
pytest installed, e.g.:

    $ pip install pytest

To run the tests, use:

    $ ./run_tests.py

For more details check <https://rdflib.readthedocs.io/en/stable/developers.html>.

Coverage
========

If ``pytest-cov`` is placed in $PYTHONPATH, it can be used to create coverage
information if the "--cov" option is supplied.

See <https://github.com/pytest-dev/pytest-cov> for details.

"""

import json
import sys

if __name__ == "__main__":
    try:
        import pytest
    except ImportError:
        print(
            """\
    Requires pytest. Try:

        $ pip install pytest

    Exiting. """,
            file=sys.stderr,
        )
        exit(1)

    finalArgs = sys.argv[1:]  # noqa: N816
    print("Running pytest with:", json.dumps(finalArgs))
    sys.exit(pytest.main(args=finalArgs))



"""
Dirt Simple Events

A Dispatcher (or a subclass of Dispatcher) stores event handlers that
are 'fired' simple event objects when interesting things happen.

Create a dispatcher:

  >>> d = Dispatcher()

Now create a handler for the event and subscribe it to the dispatcher
to handle Event events.  A handler is a simple function or method that
accepts the event as an argument:

  >>> def handler1(event): print(repr(event))
  >>> d.subscribe(Event, handler1) # doctest: +ELLIPSIS
  <rdflib.events.Dispatcher object at ...>

Now dispatch a new event into the dispatcher, and see handler1 get
fired:

  >>> d.dispatch(Event(foo='bar', data='yours', used_by='the event handlers'))
  <rdflib.events.Event ['data', 'foo', 'used_by']>
"""

from __future__ import annotations

from typing import Any, Dict, Optional

__all__ = ["Event", "Dispatcher"]


class Event:
    """
    An event is a container for attributes.  The source of an event
    creates this object, or a subclass, gives it any kind of data that
    the events handlers need to handle the event, and then calls
    notify(event).

    The target of an event registers a function to handle the event it
    is interested with subscribe().  When a sources calls
    notify(event), each subscriber to that event will be called in no
    particular order.
    """

    def __init__(self, **kw):
        self.__dict__.update(kw)

    def __repr__(self):
        attrs = sorted(self.__dict__.keys())
        return "<rdflib.events.Event %s>" % ([a for a in attrs],)


class Dispatcher:
    """
    An object that can dispatch events to a privately managed group of
    subscribers.
    """

    _dispatch_map: Optional[Dict[Any, Any]] = None

    def set_map(self, amap: Dict[Any, Any]):
        self._dispatch_map = amap
        return self

    def get_map(self):
        return self._dispatch_map

    def subscribe(self, event_type, handler):
        """Subscribe the given handler to an event_type.  Handlers
        are called in the order they are subscribed.
        """
        if self._dispatch_map is None:
            self.set_map({})
        # type error: error: Item "None" of "Optional[Dict[Any, Any]]" has no attribute "get"
        lst = self._dispatch_map.get(event_type, None)  # type: ignore[union-attr]
        if lst is None:
            lst = [handler]
        else:
            lst.append(handler)
        # type error: Unsupported target for indexed assignment ("Optional[Dict[Any, Any]]")
        self._dispatch_map[event_type] = lst  # type: ignore[index]
        return self

    def dispatch(self, event):
        """Dispatch the given event to the subscribed handlers for
        the event's type"""
        if self._dispatch_map is not None:
            lst = self._dispatch_map.get(type(event), None)
            if lst is None:
                raise ValueError("unknown event type: %s" % type(event))
            for l_ in lst:
                l_(event)



import re
from urllib.request import urlopen

import pytest

from rdflib import BNode, ConjunctiveGraph, Graph, Literal, URIRef
from test.data import BOB, CHEESE, HATES, LIKES, MICHEL, PIZZA, TAREK

HOST = "http://localhost:3031"
DB = "/db/"

# this assumes SPARQL1.1 query/update endpoints running locally at
# http://localhost:3031/db/
#
# The ConjunctiveGraph tests below require that the SPARQL endpoint renders its
# default graph as the union of all known graphs! This is incompatible with the
# endpoint behavior required by our Dataset tests in test_dataset.py, so you
# need to run a second SPARQL endpoint on a non standard port,
# e.g. fuseki started with:
# ./fuseki-server --port 3031 --memTDB --update --set tdb:unionDefaultGraph=true /db

# THIS WILL DELETE ALL DATA IN THE /db dataset

graphuri = URIRef("urn:example:graph")
othergraphuri = URIRef("urn:example:othergraph")

try:
    assert len(urlopen(HOST).read()) > 0
except Exception:
    pytest.skip(f"{HOST} is unavailable.", allow_module_level=True)


@pytest.fixture
def get_graph():
    longMessage = True  # noqa: F841
    graph = ConjunctiveGraph("SPARQLUpdateStore")

    root = HOST + DB
    graph.open((root + "sparql", root + "update"))

    # clean out the store
    for c in graph.contexts():
        c.remove((None, None, None))
        assert len(c) == 0

    yield graph

    graph.close()


def test_simple_graph(get_graph):
    graph = get_graph
    g = graph.get_context(graphuri)
    g.add((TAREK, LIKES, PIZZA))
    g.add((BOB, LIKES, PIZZA))
    g.add((BOB, LIKES, CHEESE))

    g2 = graph.get_context(othergraphuri)
    g2.add((MICHEL, LIKES, PIZZA))

    assert len(g) == 3, "graph contains 3 triples"
    assert len(g2) == 1, "other graph contains 1 triple"

    r = g.query("SELECT * WHERE { ?s <urn:example:likes> <urn:example:pizza> . }")
    assert len(list(r)) == 2, "two people like PIZZA"

    r = g.triples((None, LIKES, PIZZA))
    assert len(list(r)) == 2, "two people like PIZZA"

    # Test initBindings
    r = g.query(
        "SELECT * WHERE { ?s <urn:example:likes> <urn:example:pizza> . }",
        initBindings={"s": TAREK},
    )
    assert len(list(r)) == 1, "i was asking only about TAREK"

    r = g.triples((TAREK, LIKES, PIZZA))
    assert len(list(r)) == 1, "i was asking only about TAREK"

    r = g.triples((TAREK, LIKES, CHEESE))
    assert len(list(r)) == 0, "TAREK doesn't like CHEESE"

    g2.add((TAREK, LIKES, PIZZA))
    g.remove((TAREK, LIKES, PIZZA))
    r = g.query("SELECT * WHERE { ?s <urn:example:likes> <urn:example:pizza> . }")
    assert len(list(r)) == 1, "only BOB LIKES PIZZA"


def test_conjunctive_default(get_graph):
    graph = get_graph
    g = graph.get_context(graphuri)
    g.add((TAREK, LIKES, PIZZA))
    g2 = graph.get_context(othergraphuri)
    g2.add((BOB, LIKES, PIZZA))
    g.add((TAREK, HATES, CHEESE))

    assert 2 == len(g), "graph contains 2 triples"

    # the following are actually bad tests as they depend on your endpoint,
    # as pointed out in the sparqlstore.py code:
    #
    # For ConjunctiveGraphs, reading is done from the "default graph" Exactly
    # what this means depends on your endpoint, because SPARQL does not offer a
    # simple way to query the union of all graphs as it would be expected for a
    # ConjuntiveGraph.
    ##
    # Fuseki/TDB has a flag for specifying that the default graph
    # is the union of all graphs (tdb:unionDefaultGraph in the Fuseki config).
    assert (
        len(graph) == 3
    ), "default union graph should contain three triples but contains:\n%s" % list(
        graph
    )

    r = graph.query("SELECT * WHERE { ?s <urn:example:likes> <urn:example:pizza> . }")
    assert len(list(r)) == 2, "two people like PIZZA"

    r = graph.query(
        "SELECT * WHERE { ?s <urn:example:likes> <urn:example:pizza> . }",
        initBindings={"s": TAREK},
    )
    assert len(list(r)) == 1, "i was asking only about TAREK"

    r = graph.triples((TAREK, LIKES, PIZZA))
    assert len(list(r)) == 1, "i was asking only about TAREK"

    r = graph.triples((TAREK, LIKES, CHEESE))
    assert len(list(r)) == 0, "TAREK doesn't like CHEESE"

    g2.remove((BOB, LIKES, PIZZA))

    r = graph.query("SELECT * WHERE { ?s <urn:example:likes> <urn:example:pizza> . }")
    assert len(list(r)) == 1, "only TAREK LIKES PIZZA"


def test_u_update(get_graph):
    graph = get_graph
    graph.update(
        "INSERT DATA { GRAPH <urn:example:graph> { <urn:example:michel> <urn:example:likes> <urn:example:pizza> . } }"
    )

    g = graph.get_context(graphuri)
    assert 1 == len(g), "graph contains 1 triples"


def test_u_update_with_initns(get_graph):
    graph = get_graph
    graph.update(
        "INSERT DATA { GRAPH ns:graph { ns:michel ns:likes ns:pizza . } }",
        initNs={"ns": URIRef("urn:example:")},
    )

    g = graph.get_context(graphuri)
    assert set(g.triples((None, None, None))) == set(
        [(MICHEL, LIKES, PIZZA)]
    ), "only MICHEL LIKES PIZZA"


def test_update_with_init_bindings(get_graph):
    graph = get_graph
    graph.update(
        "INSERT { GRAPH <urn:example:graph> { ?a ?b ?c . } } WherE { }",
        initBindings={
            "a": URIRef("urn:example:michel"),
            "b": URIRef("urn:example:likes"),
            "c": URIRef("urn:example:pizza"),
        },
    )

    g = graph.get_context(graphuri)
    assert set(g.triples((None, None, None))) == set(
        [(MICHEL, LIKES, PIZZA)]
    ), "only MICHEL LIKES PIZZA"


def test_update_with_blank_node(get_graph):
    graph = get_graph
    graph.update(
        "INSERT DATA { GRAPH <urn:example:graph> { _:blankA <urn:example:type> <urn:example:Blank> } }"
    )
    g = graph.get_context(graphuri)
    for t in g.triples((None, None, None)):
        assert isinstance(t[0], BNode)
        assert t[1].n3() == "<urn:example:type>"
        assert t[2].n3() == "<urn:example:Blank>"


def test_update_w_with_blank_node_serialize_and_parse(get_graph):
    graph = get_graph
    graph.update(
        "INSERT DATA { GRAPH <urn:example:graph> { _:blankA <urn:example:type> <urn:example:Blank> } }"
    )
    g = graph.get_context(graphuri)
    string = g.serialize(format="ntriples")
    raised = False
    try:
        Graph().parse(data=string, format="ntriples")
    except Exception as e:  # noqa: F841
        raised = True
    assert raised is False, "Exception raised when parsing: " + string


def test_multiple_update_with_init_bindings(get_graph):
    graph = get_graph
    graph.update(
        "INSERT { GRAPH <urn:example:graph> { ?a ?b ?c . } } WHERE { };"
        "INSERT { GRAPH <urn:example:graph> { ?d ?b ?c . } } WHERE { }",
        initBindings={
            "a": URIRef("urn:example:michel"),
            "b": URIRef("urn:example:likes"),
            "c": URIRef("urn:example:pizza"),
            "d": URIRef("urn:example:bob"),
        },
    )

    g = graph.get_context(graphuri)
    assert set(g.triples((None, None, None))) == set(
        [(MICHEL, LIKES, PIZZA), (BOB, LIKES, PIZZA)]
    ), "MICHEL and BOB like PIZZA"


def test_named_graph_update(get_graph):
    graph = get_graph
    g = graph.get_context(graphuri)
    r1 = "INSERT DATA { <urn:example:michel> <urn:example:likes> <urn:example:pizza> }"
    g.update(r1)
    assert set(g.triples((None, None, None))) == set(
        [(MICHEL, LIKES, PIZZA)]
    ), "only MICHEL LIKES PIZZA"

    r2 = (
        "DELETE { <urn:example:michel> <urn:example:likes> <urn:example:pizza> } "
        + "INSERT { <urn:example:bob> <urn:example:likes> <urn:example:pizza> } WHERE {}"
    )
    g.update(r2)
    assert set(g.triples((None, None, None))) == set(
        [(BOB, LIKES, PIZZA)]
    ), "only BOB LIKES PIZZA"

    says = URIRef("urn:says")

    # Strings with unbalanced curly braces
    tricky_strs = ["With an unbalanced curly brace %s " % brace for brace in ["{", "}"]]
    for tricky_str in tricky_strs:
        r3 = (
            """INSERT { ?b <urn:says> "%s" }
        WHERE { ?b <urn:example:likes> <urn:example:pizza>} """
            % tricky_str
        )
        g.update(r3)

    values = set()
    for v in g.objects(BOB, says):
        values.add(str(v))
    assert values == set(tricky_strs)

    # Complicated Strings
    r4strings = []
    r4strings.append(r'''"1: adfk { ' \\\" \" { "''')
    r4strings.append(r'''"2: adfk } <foo> #éï \\"''')

    r4strings.append(r"""'3: adfk { " \\\' \' { '""")
    r4strings.append(r"""'4: adfk } <foo> #éï \\'""")

    r4strings.append(r'''"""5: adfk { ' \\\" \" { """''')
    r4strings.append(r'''"""6: adfk } <foo> #éï \\"""''')
    r4strings.append('"""7: ad adsfj \n { \n sadfj"""')

    r4strings.append(r"""'''8: adfk { " \\\' \' { '''""")
    r4strings.append(r"""'''9: adfk } <foo> #éï \\'''""")
    r4strings.append("'''10: ad adsfj \n { \n sadfj'''")

    r4 = "\n".join(
        ["INSERT DATA { <urn:example:michel> <urn:says> %s } ;" % s for s in r4strings]
    )
    g.update(r4)
    values = set()
    for v in g.objects(MICHEL, says):
        values.add(str(v))
    assert values == set(
        [
            re.sub(
                r"\\(.)",
                r"\1",
                re.sub(r"^'''|'''$|^'|'$|" + r'^"""|"""$|^"|"$', r"", s),
            )
            for s in r4strings
        ]
    )

    # IRI Containing ' or #
    # The fragment identifier must not be misinterpreted as a comment
    # (commenting out the end of the block).
    # The ' must not be interpreted as the start of a string, causing the }
    # in the literal to be identified as the end of the block.
    r5 = """INSERT DATA { <urn:example:michel> <urn:example:hates> <urn:example:foo'bar?baz;a=1&b=2#fragment>, "'}" }"""

    g.update(r5)
    values = set()
    for v in g.objects(MICHEL, HATES):
        values.add(str(v))
    assert values == set(["urn:example:foo'bar?baz;a=1&b=2#fragment", "'}"])

    # Comments
    r6 = """
        INSERT DATA {
            <urn:example:bob> <urn:example:hates> <urn:example:bob> . # No closing brace: }
            <urn:example:bob> <urn:example:hates> <urn:example:michel>.
        }
    #Final { } comment"""

    g.update(r6)
    values = set()
    for v in g.objects(BOB, HATES):
        values.add(v)
    assert values == set([BOB, MICHEL])


def test_named_graph_update_with_init_bindings(get_graph):
    graph = get_graph
    g = graph.get_context(graphuri)
    r = "INSERT { ?a ?b ?c } WHERE {}"
    g.update(r, initBindings={"a": MICHEL, "b": LIKES, "c": PIZZA})
    assert set(g.triples((None, None, None))) == set(
        [(MICHEL, LIKES, PIZZA)]
    ), "only MICHEL LIKES PIZZA"


def test_empty_named_graph(get_graph):
    graph = get_graph
    empty_graph_iri = "urn:empty-graph-1"
    graph.update("CREATE GRAPH <%s>" % empty_graph_iri)
    named_graphs = [
        str(r[0]) for r in graph.query("SELECT ?name WHERE { GRAPH ?name {} }")
    ]
    # Some SPARQL endpoint backends (like TDB) are not able to find empty named graphs
    # (at least with this query)
    if empty_graph_iri in named_graphs:
        assert empty_graph_iri in [str(g.identifier) for g in graph.contexts()]


def test_empty_literal(get_graph):
    graph = get_graph
    # test for https://github.com/RDFLib/rdflib/issues/457
    # also see test_issue457.py which is sparql store independent!
    g = graph.get_context(graphuri)
    g.add(
        (
            URIRef("http://example.com/s"),
            URIRef("http://example.com/p"),
            Literal(""),
        )
    )

    o = tuple(g)[0][2]
    assert o == Literal(""), repr(o)



from __future__ import annotations

from codecs import getreader
from enum import Enum
from typing import TYPE_CHECKING, Any, MutableMapping, Optional, Union

from rdflib.exceptions import ParserError as ParseError
from rdflib.graph import Dataset
from rdflib.parser import InputSource
from rdflib.plugins.parsers.nquads import NQuadsParser

# Build up from the NTriples parser:
from rdflib.plugins.parsers.ntriples import r_nodeid, r_tail, r_uriref, r_wspace
from rdflib.term import BNode, URIRef

if TYPE_CHECKING:
    import typing_extensions as te

__all__ = ["RDFPatchParser", "Operation"]

_BNodeContextType = MutableMapping[str, BNode]


class Operation(Enum):
    """
    Enum of RDF Patch operations.

    Operations:
    - `AddTripleOrQuad` (A): Adds a triple or quad.
    - `DeleteTripleOrQuad` (D): Deletes a triple or quad.
    - `AddPrefix` (PA): Adds a prefix.
    - `DeletePrefix` (PD): Deletes a prefix.
    - `TransactionStart` (TX): Starts a transaction.
    - `TransactionCommit` (TC): Commits a transaction.
    - `TransactionAbort` (TA): Aborts a transaction.
    - `Header` (H): Specifies a header.
    """

    AddTripleOrQuad = "A"
    DeleteTripleOrQuad = "D"
    AddPrefix = "PA"
    DeletePrefix = "PD"
    TransactionStart = "TX"
    TransactionCommit = "TC"
    TransactionAbort = "TA"
    Header = "H"


class RDFPatchParser(NQuadsParser):
    def parse(  # type: ignore[override]
        self,
        inputsource: InputSource,
        sink: Dataset,
        bnode_context: Optional[_BNodeContextType] = None,
        skolemize: bool = False,
        **kwargs: Any,
    ) -> Dataset:
        """
        Parse inputsource as an RDF Patch file.

        :type inputsource: `rdflib.parser.InputSource`
        :param inputsource: the source of RDF Patch formatted data
        :type sink: `rdflib.graph.Dataset`
        :param sink: where to send parsed data
        :type bnode_context: `dict`, optional
        :param bnode_context: a dict mapping blank node identifiers to `~rdflib.term.BNode` instances.
                              See `.W3CNTriplesParser.parse`
        """
        assert sink.store.context_aware, (
            "RDFPatchParser must be given" " a context aware store."
        )
        # type error: Incompatible types in assignment (expression has type "ConjunctiveGraph", base class "W3CNTriplesParser" defined the type as "Union[DummySink, NTGraphSink]")
        self.sink: Dataset = Dataset(store=sink.store)
        self.skolemize = skolemize

        source = inputsource.getCharacterStream()
        if not source:
            source = inputsource.getByteStream()
            source = getreader("utf-8")(source)

        if not hasattr(source, "read"):
            raise ParseError("Item to parse must be a file-like object.")

        self.file = source
        self.buffer = ""
        while True:
            self.line = __line = self.readline()
            if self.line is None:
                break
            try:
                self.parsepatch(bnode_context)
            except ParseError as msg:
                raise ParseError("Invalid line (%s):\n%r" % (msg, __line))
        return self.sink

    def parsepatch(self, bnode_context: Optional[_BNodeContextType] = None) -> None:
        self.eat(r_wspace)
        #  From spec: "No comments should be included (comments start # and run to end
        #  of line)."
        if (not self.line) or self.line.startswith("#"):
            return  # The line is empty or a comment

        # if header, transaction, skip
        operation = self.operation()
        self.eat(r_wspace)

        if operation in [Operation.AddTripleOrQuad, Operation.DeleteTripleOrQuad]:
            self.add_or_remove_triple_or_quad(operation, bnode_context)
        elif operation == Operation.AddPrefix:
            self.add_prefix()
        elif operation == Operation.DeletePrefix:
            self.delete_prefix()

    def add_or_remove_triple_or_quad(
        self, operation, bnode_context: Optional[_BNodeContextType] = None
    ) -> None:
        self.eat(r_wspace)
        if (not self.line) or self.line.startswith("#"):
            return  # The line is empty or a comment

        subject = self.labeled_bnode() or self.subject(bnode_context)
        self.eat(r_wspace)

        predicate = self.predicate()
        self.eat(r_wspace)

        obj = self.labeled_bnode() or self.object(bnode_context)
        self.eat(r_wspace)

        context = self.labeled_bnode() or self.uriref() or self.nodeid(bnode_context)
        self.eat(r_tail)

        if self.line:
            raise ParseError("Trailing garbage")
        # Must have a context aware store - add on a normal Graph
        # discards anything where the ctx != graph.identifier
        if operation == Operation.AddTripleOrQuad:
            if context:
                self.sink.get_context(context).add((subject, predicate, obj))
            else:
                self.sink.default_context.add((subject, predicate, obj))
        elif operation == Operation.DeleteTripleOrQuad:
            if context:
                self.sink.get_context(context).remove((subject, predicate, obj))
            else:
                self.sink.default_context.remove((subject, predicate, obj))

    def add_prefix(self):
        # Extract prefix and URI from the line
        prefix, ns, _ = self.line.replace('"', "").replace("'", "").split(" ")  # type: ignore[union-attr]
        ns_stripped = ns.strip("<>")
        self.sink.bind(prefix, ns_stripped)

    def delete_prefix(self):
        prefix, _, _ = self.line.replace('"', "").replace("'", "").split(" ")  # type: ignore[union-attr]
        self.sink.namespace_manager.bind(prefix, None, replace=True)

    def operation(self) -> Operation:
        for op in Operation:
            if self.line.startswith(op.value):  # type: ignore[union-attr]
                self.eat_op(op.value)
                return op
        raise ValueError(
            f'Invalid or no Operation found in line: "{self.line}". Valid Operations '
            f"codes are {', '.join([op.value for op in Operation])}"
        )

    def eat_op(self, op: str) -> None:
        self.line = self.line.lstrip(op)  # type: ignore[union-attr]

    def nodeid(
        self, bnode_context: Optional[_BNodeContextType] = None
    ) -> Union[te.Literal[False], BNode, URIRef]:
        if self.peek("_"):
            return BNode(self.eat(r_nodeid).group(1))
        return False

    def labeled_bnode(self):
        if self.peek("<_"):
            plain_uri = self.eat(r_uriref).group(1)
            bnode_id = r_nodeid.match(plain_uri).group(1)  # type: ignore[union-attr]
            return BNode(bnode_id)
        return False



from rdflib import BNode, Graph
from rdflib.compare import isomorphic
from test.utils.namespace import EGDO


def test_no_spurious_semicolon():
    sparql = """
      PREFIX : <http://example.org/>
      CONSTRUCT {
        :a :b :c ; :d :e .
      } WHERE {}
    """
    expected = Graph()
    expected.addN(
        t + (expected,)
        for t in [
            (EGDO.a, EGDO.b, EGDO.c),
            (EGDO.a, EGDO.d, EGDO.e),
        ]
    )
    got = Graph().query(sparql).graph
    assert isomorphic(got, expected), got.serialize(format="turtle")


def test_one_spurious_semicolon():
    sparql = """
      PREFIX : <http://example.org/>
      CONSTRUCT {
        :a :b :c ; :d :e ; .
      } WHERE {}
    """
    expected = Graph()
    expected.addN(
        t + (expected,)
        for t in [
            (EGDO.a, EGDO.b, EGDO.c),
            (EGDO.a, EGDO.d, EGDO.e),
        ]
    )
    got = Graph().query(sparql).graph
    assert isomorphic(got, expected), got.serialize(format="turtle")


def test_one_spurious_semicolon_no_perdiod():
    sparql = """
      PREFIX : <http://example.org/>
      CONSTRUCT {
        :a :b :c ; :d :e ;
      } WHERE {}
    """
    expected = Graph()
    expected.addN(
        t + (expected,)
        for t in [
            (EGDO.a, EGDO.b, EGDO.c),
            (EGDO.a, EGDO.d, EGDO.e),
        ]
    )
    got = Graph().query(sparql).graph
    assert isomorphic(got, expected), got.serialize(format="turtle")


def test_two_spurious_semicolons_no_period():
    sparql = """
      PREFIX : <http://example.org/>
      CONSTRUCT {
        :a :b :c ; :d :e ; ;
      } WHERE {}
    """
    expected = Graph()
    expected.addN(
        t + (expected,)
        for t in [
            (EGDO.a, EGDO.b, EGDO.c),
            (EGDO.a, EGDO.d, EGDO.e),
        ]
    )
    got = Graph().query(sparql).graph
    assert isomorphic(got, expected), got.serialize(format="turtle")


def test_one_spurious_semicolons_bnode():
    sparql = """
      PREFIX : <http://example.org/>
      CONSTRUCT {
        [ :b :c ; :d :e ; ]
      } WHERE {}
    """
    expected = Graph()
    expected.addN(
        t + (expected,)
        for t in [
            (BNode("a"), EGDO.b, EGDO.c),
            (BNode("a"), EGDO.d, EGDO.e),
        ]
    )
    got = Graph().query(sparql).graph
    assert isomorphic(got, expected), got.serialize(format="turtle")


def test_pathological():
    """
    This test did not raise an exception,
    but generated a graph completely different from the expected one.
    """
    sparql = """
      PREFIX : <http://example.org/>
      CONSTRUCT {
        :a :b :c ; ;
           :d :e ; ;
           :f :g ;
      } WHERE {}
    """
    expected = Graph()
    expected.addN(
        t + (expected,)
        for t in [
            (EGDO.a, EGDO.b, EGDO.c),
            (EGDO.a, EGDO.d, EGDO.e),
            (EGDO.a, EGDO.f, EGDO.g),
        ]
    )
    got = Graph().query(sparql).graph
    assert isomorphic(got, expected), got.serialize(format="turtle")


def test_mixing_spurious_semicolons_and_commas():
    sparql = """
      PREFIX : <http://example.org/>
      CONSTRUCT {
        :a :b :c ; ;
           :d :e, :f
      } WHERE {}
    """
    expected = Graph()
    expected.addN(
        t + (expected,)
        for t in [
            (EGDO.a, EGDO.b, EGDO.c),
            (EGDO.a, EGDO.d, EGDO.e),
            (EGDO.a, EGDO.d, EGDO.f),
        ]
    )
    got = Graph().query(sparql).graph
    assert isomorphic(got, expected), got.serialize(format="turtle")



"""
RDFLib Graphs (and Resources) can be "sliced" with [] syntax

This is a short-hand for iterating over triples.

Combined with SPARQL paths (see ``foafpaths.py``) - quite complex queries
can be realised.

See :meth:`rdflib.graph.Graph.__getitem__` for details
"""

from pathlib import Path

from rdflib import RDF, Graph
from rdflib.namespace import FOAF

EXAMPLES_DIR = Path(__file__).parent


if __name__ == "__main__":
    graph = Graph()
    graph.parse(f"{EXAMPLES_DIR / 'foaf.n3'}", format="n3")

    for person in graph[: RDF.type : FOAF.Person]:  # type: ignore[misc]
        friends = list(graph[person : FOAF.knows * "+" / FOAF.name])  # type: ignore[operator]
        if friends:
            print(f"{graph.value(person, FOAF.name)}'s circle of friends:")
            for name in friends:
                print(name)



from __future__ import annotations

import collections
import email.message
import enum
import random
from contextlib import contextmanager
from http.server import BaseHTTPRequestHandler, HTTPServer
from threading import Thread
from typing import (
    Dict,
    Iterable,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from urllib.parse import ParseResult

from test.utils.wildcard import EQ_WILDCARD

__all__: List[str] = []

HeadersT = Union[Dict[str, List[str]], Iterable[Tuple[str, str]]]
PathQueryT = Dict[str, List[str]]


def header_items(headers: HeadersT) -> Iterable[Tuple[str, str]]:
    if isinstance(headers, collections.abc.Mapping):
        for header, value in headers.items():
            if isinstance(value, list):
                for item in value:
                    yield header, item
    else:
        yield from headers


def apply_headers_to(headers: HeadersT, handler: BaseHTTPRequestHandler) -> None:
    for header, value in header_items(headers):
        handler.send_header(header, value)
    # handler.end_headers()


class MethodName(str, enum.Enum):
    CONNECT = enum.auto()
    DELETE = enum.auto()
    GET = enum.auto()
    HEAD = enum.auto()
    OPTIONS = enum.auto()
    PATCH = enum.auto()
    POST = enum.auto()
    PUT = enum.auto()
    TRACE = enum.auto()


class MockHTTPRequest(NamedTuple):
    method: MethodName
    path: str
    parsed_path: ParseResult
    path_query: PathQueryT
    headers: email.message.Message
    body: Optional[bytes]


MOCK_HTTP_REQUEST_WILDCARD = MockHTTPRequest(
    EQ_WILDCARD, EQ_WILDCARD, EQ_WILDCARD, EQ_WILDCARD, EQ_WILDCARD, EQ_WILDCARD
)
"""
This object should be equal to any `MockHTTPRequest` object.
"""


class MockHTTPResponse(NamedTuple):
    status_code: int
    reason_phrase: str
    body: bytes
    headers: HeadersT


def get_random_ip(ip_prefix: Optional[List[str]] = None) -> str:
    if ip_prefix is None:
        parts = ["127"]
    for _ in range(4 - len(parts)):
        parts.append(f"{random.randint(0, 255)}")
    return ".".join(parts)


@contextmanager
def ctx_http_handler(
    handler: Type[BaseHTTPRequestHandler], host: Optional[str] = "127.0.0.1"
) -> Iterator[HTTPServer]:
    host = get_random_ip() if host is None else host
    server = HTTPServer((host, 0), handler)
    with ctx_http_server(server) as server:
        yield server


HTTPServerT = TypeVar("HTTPServerT", bound=HTTPServer)


@contextmanager
def ctx_http_server(server: HTTPServerT) -> Iterator[HTTPServerT]:
    server_thread = Thread(target=server.serve_forever)
    server_thread.daemon = True
    server_thread.start()
    yield server
    server.shutdown()
    server.socket.close()
    server_thread.join()


def headers_as_message(headers: HeadersT) -> email.message.Message:
    message = email.message.Message()
    for header, value in header_items(headers):
        # This will append the value to any existing values for the header
        # instead of replacing it.
        message[header] = value
    return message



from rdflib.graph import Dataset, Graph
from rdflib.term import URIRef

dataset = Dataset(default_union=True)
# Adding into default graph
dataset.add((URIRef("urn:s0"), URIRef("urn:p0"), URIRef("urn:o0")))
# Adding into named graphs
dataset.add(
    (
        URIRef("urn:s1"),
        URIRef("urn:p1"),
        URIRef("urn:o1"),
        Graph(identifier=URIRef("urn:g1")),
    )
)

dataset.add(
    (
        URIRef("urn:s2"),
        URIRef("urn:p2"),
        URIRef("urn:o2"),
        Graph(identifier=URIRef("urn:g2")),
    )
)

dataset.add(
    (
        URIRef("urn:s3"),
        URIRef("urn:p3"),
        URIRef("urn:o3"),
        Graph(identifier=URIRef("urn:g3")),
    )
)


# Test implicit inclusive dataset
# The query's default graph should contain a merge of all graphs:
# The service's default graph + all the service's named graphs
def test_inclusive():
    query = """
        SELECT ?s ?p ?o
        WHERE {?s ?p ?o}
        ORDER BY ?s
    """
    results = list(dataset.query(query))
    assert results == [
        (URIRef("urn:s0"), URIRef("urn:p0"), URIRef("urn:o0")),
        (URIRef("urn:s1"), URIRef("urn:p1"), URIRef("urn:o1")),
        (URIRef("urn:s2"), URIRef("urn:p2"), URIRef("urn:o2")),
        (URIRef("urn:s3"), URIRef("urn:p3"), URIRef("urn:o3")),
    ]


# Test explicit default graph with inclusive dataset
def test_default_from_1():
    query = """
        SELECT ?s ?p ?o
        FROM <urn:g1>
        WHERE {?s ?p ?o}
    """
    results = list(dataset.query(query))
    assert results == [(URIRef("urn:s1"), URIRef("urn:p1"), URIRef("urn:o1"))]


# test that we include more than one graph into the default graph
def test_default_from_2():
    query = """
        SELECT ?s ?p ?o
        FROM <urn:g1>
        FROM <urn:g2>
        WHERE {?s ?p ?o}
        ORDER BY ?s
    """
    results = list(dataset.query(query))
    assert results == [
        (URIRef("urn:s1"), URIRef("urn:p1"), URIRef("urn:o1")),
        (URIRef("urn:s2"), URIRef("urn:p2"), URIRef("urn:o2")),
    ]


# Since there is a FROM clause, we consider RDF dataset explicit
# Thus if FROM NAMED is not defined, named graph is considered empty set
def test_named_from():
    query = """
        SELECT ?s ?p ?o
        FROM <urn:g1>
        WHERE {
            graph ?g {?s ?p ?o}
        } ORDER BY ?s
    """
    results = list(dataset.query(query))
    assert results == [], "no result expected"


# Test explicit named graphs with inclusive dataset
def test_named_from_named_1():
    query = """
        SELECT ?g ?s ?p ?o
        FROM NAMED <urn:g1>
        WHERE {
            graph ?g {?s ?p ?o}
        }
    """
    results = list(dataset.query(query))
    assert results == [
        (URIRef("urn:g1"), URIRef("urn:s1"), URIRef("urn:p1"), URIRef("urn:o1"))
    ]


# test that we include more than one graph into the named graphs
def test_named_from_named_2():
    query = """
    SELECT ?g ?s ?p ?o
    FROM NAMED <urn:g1>
    FROM NAMED <urn:g2>
    WHERE {
        graph ?g {?s ?p ?o}
    } ORDER BY ?g
    """
    results = list(dataset.query(query))
    assert results == [
        (URIRef("urn:g1"), URIRef("urn:s1"), URIRef("urn:p1"), URIRef("urn:o1")),
        (URIRef("urn:g2"), URIRef("urn:s2"), URIRef("urn:p2"), URIRef("urn:o2")),
    ]


# Since there is a FROM NAMED clause, we consider RDF dataset explicit
# Thus if FROM is not defined, default graph is considered empty
def test_default_from_named():
    results = list(
        dataset.query("SELECT ?g ?s ?p ?o FROM NAMED <urn:g1> WHERE {?s ?p ?o}")
    )
    assert results == [], "no result expected"


def test_from_and_from_named():
    query = """
        SELECT ?g ?s ?p ?o
        FROM <urn:g1>
        FROM NAMED <urn:g2>
        WHERE {
            {?s ?p ?o}
            UNION {graph ?g {?s ?p ?o}}
        } ORDER BY ?s
    """
    results = list(dataset.query(query))
    assert results == [
        (None, URIRef("urn:s1"), URIRef("urn:p1"), URIRef("urn:o1")),
        (URIRef("urn:g2"), URIRef("urn:s2"), URIRef("urn:p2"), URIRef("urn:o2")),
    ]



from __future__ import annotations

import enum
import logging
import posixpath
from dataclasses import dataclass, field
from functools import lru_cache
from http.server import BaseHTTPRequestHandler, HTTPServer
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Type
from urllib.parse import parse_qs, urljoin, urlparse
from uuid import uuid4

from test.utils.http import HeadersT, MethodName, MockHTTPRequest, apply_headers_to

__all__: List[str] = [
    "LocationType",
    "ProtoResource",
    "Resource",
    "ProtoRedirectResource",
    "ProtoFileResource",
    "RedirectResource",
    "FileResource",
    "HTTPFileInfo",
    "HTTPFileServer",
]


class LocationType(enum.Enum):
    RELATIVE_PATH = enum.auto()
    ABSOLUTE_PATH = enum.auto()
    URL = enum.auto()


@dataclass(
    frozen=True,
)
class ProtoResource:
    headers: HeadersT


@dataclass(frozen=True)
class Resource(ProtoResource):
    url_path: str
    url: str


@dataclass(frozen=True)
class ProtoRedirectResource(ProtoResource):
    status: int
    location_type: LocationType


@dataclass(frozen=True)
class ProtoFileResource(ProtoResource):
    file_path: Path


@dataclass(frozen=True)
class RedirectResource(Resource, ProtoRedirectResource):
    location: str


@dataclass(frozen=True)
class FileResource(Resource, ProtoFileResource):
    pass


@dataclass(frozen=True)
class HTTPFileInfo:
    """
    Information about a file served by the HTTPFileServerRequestHandler.

    :param request_url: The URL that should be requested to get the file.
    :param effective_url: The URL that the file will be served from after
        redirects.
    :param redirects: A sequence of redirects that will be given to the client
        if it uses the ``request_url``. This sequence will terminate in the
        ``effective_url``.
    """

    # request_url: str
    # effective_url: str
    file: FileResource
    redirects: Sequence[RedirectResource] = field(default_factory=list)

    @property
    def path(self) -> Path:
        return self.file.file_path

    @property
    def request_url(self) -> str:
        """
        The URL that should be requested to get the file.
        """
        if self.redirects:
            return self.redirects[0].url
        else:
            return self.file.url

    @property
    def effective_url(self) -> str:
        """
        The URL that the file will be served from after
        redirects.
        """
        return self.file.url


class HTTPFileServer(HTTPServer):
    def __init__(
        self,
        server_address: tuple[str, int],
        bind_and_activate: bool = True,
    ) -> None:
        self._resources: Dict[str, Resource] = {}
        self.Handler = self.make_handler()
        super().__init__(server_address, self.Handler, bind_and_activate)

    @property
    def url(self) -> str:
        (host, port) = self.server_address
        if isinstance(host, (bytes, bytearray)):
            host = host.decode("utf-8")
        return f"http://{host}:{port}"

    @lru_cache(maxsize=1024)
    def add_file_with_caching(
        self,
        proto_file: ProtoFileResource,
        proto_redirects: Optional[Sequence[ProtoRedirectResource]] = None,
        suffix: str = "",
    ) -> HTTPFileInfo:
        return self.add_file(proto_file, proto_redirects, suffix)

    def add_file(
        self,
        proto_file: ProtoFileResource,
        proto_redirects: Optional[Sequence[ProtoRedirectResource]] = None,
        suffix: str = "",
    ) -> HTTPFileInfo:
        url_path = f"/file/{uuid4().hex}{suffix}"
        url = urljoin(self.url, url_path)
        file_resource = FileResource(
            url_path=url_path,
            url=url,
            file_path=proto_file.file_path,
            headers=proto_file.headers,
        )
        self._resources[url_path] = file_resource

        if proto_redirects is None:
            proto_redirects = []

        redirects: List[RedirectResource] = []
        for proto_redirect in reversed(proto_redirects):
            redirect_url_path = f"/redirect/{uuid4().hex}{suffix}"
            if proto_redirect.location_type == LocationType.URL:
                location = url
            elif proto_redirect.location_type == LocationType.ABSOLUTE_PATH:
                location = url_path
            elif proto_redirect.location_type == LocationType.RELATIVE_PATH:
                location = posixpath.relpath(url_path, redirect_url_path)
            else:
                raise ValueError(
                    f"unsupported location_type={proto_redirect.location_type}"
                )
            url_path = redirect_url_path
            url = urljoin(self.url, url_path)
            redirect_resource = RedirectResource(
                url_path=url_path,
                url=url,
                status=proto_redirect.status,
                location_type=proto_redirect.location_type,
                location=location,
                headers=proto_redirect.headers,
            )
            self._resources[url_path] = redirect_resource

        file_info = HTTPFileInfo(file_resource, redirects)
        return file_info

    def make_handler(self) -> Type[BaseHTTPRequestHandler]:
        class Handler(BaseHTTPRequestHandler):
            server: HTTPFileServer

            def do_GET(self) -> None:  # noqa: N802
                parsed_path = urlparse(self.path)
                path_query = parse_qs(parsed_path.query)
                body = None
                content_length = self.headers.get("Content-Length")
                if content_length is not None:
                    body = self.rfile.read(int(content_length))
                method_name = MethodName.GET
                request = MockHTTPRequest(
                    method_name,
                    self.path,
                    parsed_path,
                    path_query,
                    self.headers,
                    body,
                )
                logging.debug("handling %s request: %s", method_name, request)
                logging.debug("headers %s", request.headers)

                resource_path = parsed_path.path
                if resource_path not in self.server._resources:
                    self.send_error(404, "File not found")
                    return

                resource = self.server._resources[resource_path]
                if isinstance(resource, FileResource):
                    self.send_response(200)
                elif isinstance(resource, RedirectResource):
                    self.send_response(resource.status)
                    self.send_header("Location", resource.location)
                apply_headers_to(resource.headers, self)

                self.end_headers()

                if isinstance(resource, FileResource):
                    with resource.file_path.open("rb") as f:
                        self.wfile.write(f.read())
                        self.wfile.flush()
                return

        Handler.server = self

        return Handler



"""
Turtle RDF graph serializer for RDFLib.
See <http://www.w3.org/TeamSubmission/turtle/> for syntax specification.
"""

from __future__ import annotations

from collections import defaultdict
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    DefaultDict,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
)

from rdflib.exceptions import Error
from rdflib.graph import Graph
from rdflib.namespace import RDF, RDFS
from rdflib.serializer import Serializer
from rdflib.term import BNode, Literal, Node, URIRef

if TYPE_CHECKING:
    from rdflib.graph import _PredicateType, _SubjectType, _TripleType

__all__ = ["RecursiveSerializer", "TurtleSerializer"]


class RecursiveSerializer(Serializer):
    topClasses = [RDFS.Class]
    predicateOrder = [RDF.type, RDFS.label]
    maxDepth = 10
    indentString = "  "
    roundtrip_prefixes: Tuple[Any, ...] = ()

    def __init__(self, store: Graph):
        super(RecursiveSerializer, self).__init__(store)
        self.stream: Optional[IO[bytes]] = None
        self.reset()

    def addNamespace(self, prefix: str, uri: URIRef) -> None:
        if prefix in self.namespaces and self.namespaces[prefix] != uri:
            raise Exception(
                "Trying to override namespace prefix %s => %s, but it's already bound to %s"
                % (prefix, uri, self.namespaces[prefix])
            )
        self.namespaces[prefix] = uri

    def checkSubject(self, subject: _SubjectType) -> bool:
        """Check to see if the subject should be serialized yet"""
        if (
            (self.isDone(subject))
            or (subject not in self._subjects)
            or ((subject in self._topLevels) and (self.depth > 1))
            or (isinstance(subject, URIRef) and (self.depth >= self.maxDepth))
        ):
            return False
        return True

    def isDone(self, subject: _SubjectType) -> bool:
        """Return true if subject is serialized"""
        return subject in self._serialized

    def orderSubjects(self) -> List[_SubjectType]:
        seen: Dict[_SubjectType, bool] = {}
        subjects: List[_SubjectType] = []

        for classURI in self.topClasses:
            members = list(self.store.subjects(RDF.type, classURI))
            members.sort()

            subjects.extend(members)
            for member in members:
                self._topLevels[member] = True
                seen[member] = True

        recursable = [
            (isinstance(subject, BNode), self._references[subject], subject)
            for subject in self._subjects
            if subject not in seen
        ]

        recursable.sort()
        subjects.extend([subject for (isbnode, refs, subject) in recursable])

        return subjects

    def preprocess(self) -> None:
        for triple in self.store.triples((None, None, None)):
            self.preprocessTriple(triple)

    def preprocessTriple(self, spo: _TripleType) -> None:
        s, p, o = spo
        self._references[o] += 1
        self._subjects[s] = True

    def reset(self) -> None:
        self.depth = 0
        # Typed none because nothing is using it ...
        self.lists: Dict[None, None] = {}
        self.namespaces: Dict[str, URIRef] = {}
        self._references: DefaultDict[Node, int] = defaultdict(int)
        self._serialized: Dict[_SubjectType, bool] = {}
        self._subjects: Dict[_SubjectType, bool] = {}
        self._topLevels: Dict[_SubjectType, bool] = {}

        if self.roundtrip_prefixes:
            if hasattr(self.roundtrip_prefixes, "__iter__"):
                for prefix, ns in self.store.namespaces():
                    if prefix in self.roundtrip_prefixes:
                        self.addNamespace(prefix, ns)
            else:
                for prefix, ns in self.store.namespaces():
                    self.addNamespace(prefix, ns)

    def buildPredicateHash(
        self, subject: _SubjectType
    ) -> Mapping[_PredicateType, List[Node]]:
        """
        Build a hash key by predicate to a list of objects for the given
        subject
        """
        properties: Dict[_PredicateType, List[Node]] = {}
        for s, p, o in self.store.triples((subject, None, None)):
            oList = properties.get(p, [])
            oList.append(o)
            properties[p] = oList
        return properties

    def sortProperties(
        self, properties: Mapping[_PredicateType, List[Node]]
    ) -> List[_PredicateType]:
        """Take a hash from predicate uris to lists of values.
        Sort the lists of values.  Return a sorted list of properties."""
        # Sort object lists
        for prop, objects in properties.items():
            objects.sort()

        # Make sorted list of properties
        propList: List[_PredicateType] = []
        seen: Dict[_PredicateType, bool] = {}
        for prop in self.predicateOrder:
            if (prop in properties) and (prop not in seen):
                propList.append(prop)
                seen[prop] = True
        props = list(properties.keys())
        props.sort()
        for prop in props:
            if prop not in seen:
                propList.append(prop)
                seen[prop] = True
        return propList

    def subjectDone(self, subject: _SubjectType) -> None:
        """Mark a subject as done."""
        self._serialized[subject] = True

    def indent(self, modifier: int = 0) -> str:
        """Returns indent string multiplied by the depth"""
        return (self.depth + modifier) * self.indentString

    def write(self, text: str) -> None:
        """Write text in given encoding."""
        # type error: Item "None" of "Optional[IO[bytes]]" has no attribute "write"
        self.stream.write(text.encode(self.encoding, "replace"))  # type: ignore[union-attr]


SUBJECT = 0
VERB = 1
OBJECT = 2

_GEN_QNAME_FOR_DT = False
_SPACIOUS_OUTPUT = False


class TurtleSerializer(RecursiveSerializer):
    short_name = "turtle"
    indentString = "    "

    def __init__(self, store: Graph):
        self._ns_rewrite: Dict[str, str] = {}
        super(TurtleSerializer, self).__init__(store)
        self.keywords: Dict[Node, str] = {RDF.type: "a"}
        self.reset()
        self.stream = None
        self._spacious = _SPACIOUS_OUTPUT

    # type error: Return type "str" of "addNamespace" incompatible with return type "None" in supertype "RecursiveSerializer"
    def addNamespace(self, prefix: str, namespace: URIRef) -> str:  # type: ignore[override]
        # Turtle does not support prefix that start with _
        # if they occur in the graph, rewrite to p_blah
        # this is more complicated since we need to make sure p_blah
        # does not already exist. And we register namespaces as we go, i.e.
        # we may first see a triple with prefix _9 - rewrite it to p_9
        # and then later find a triple with a "real" p_9 prefix

        # so we need to keep track of ns rewrites we made so far.

        if (prefix > "" and prefix[0] == "_") or self.namespaces.get(
            prefix, namespace
        ) != namespace:
            if prefix not in self._ns_rewrite:
                p = "p" + prefix
                while p in self.namespaces:
                    p = "p" + p
                self._ns_rewrite[prefix] = p

            prefix = self._ns_rewrite.get(prefix, prefix)

        super(TurtleSerializer, self).addNamespace(prefix, namespace)
        return prefix

    def reset(self) -> None:
        super(TurtleSerializer, self).reset()
        # typing as Dict[None, None] because nothing seems to be using it
        self._shortNames: Dict[None, None] = {}
        self._started = False
        self._ns_rewrite = {}

    def serialize(
        self,
        stream: IO[bytes],
        base: Optional[str] = None,
        encoding: Optional[str] = None,
        spacious: Optional[bool] = None,
        **args: Any,
    ) -> None:
        self.reset()
        self.stream = stream
        # if base is given here, use that, if not and a base is set for the graph use that
        if base is not None:
            self.base = base
        elif self.store.base is not None:
            self.base = self.store.base

        if spacious is not None:
            self._spacious = spacious

        self.preprocess()
        subjects_list = self.orderSubjects()

        self.startDocument()

        firstTime = True
        for subject in subjects_list:
            if self.isDone(subject):
                continue
            if firstTime:
                firstTime = False
            if self.statement(subject) and not firstTime:
                self.write("\n")

        self.endDocument()
        stream.write("\n".encode("latin-1"))

        self.base = None

    def preprocessTriple(self, triple: _TripleType) -> None:
        super(TurtleSerializer, self).preprocessTriple(triple)
        for i, node in enumerate(triple):
            if i == VERB and node in self.keywords:
                # predicate is a keyword
                continue
            # Don't use generated prefixes for subjects and objects
            self.getQName(node, gen_prefix=(i == VERB))
            if isinstance(node, Literal) and node.datatype:
                self.getQName(node.datatype, gen_prefix=_GEN_QNAME_FOR_DT)
        p = triple[1]
        if isinstance(p, BNode):  # hmm - when is P ever a bnode?
            self._references[p] += 1

    # TODO: Rename to get_pname
    def getQName(self, uri: Node, gen_prefix: bool = True) -> Optional[str]:
        if not isinstance(uri, URIRef):
            return None

        parts = None

        try:
            parts = self.store.compute_qname(uri, generate=gen_prefix)
        except Exception:
            # is the uri a namespace in itself?
            pfx = self.store.store.prefix(uri)

            if pfx is not None:
                parts = (pfx, uri, "")
            else:
                # nothing worked
                return None

        prefix, namespace, local = parts

        local = local.replace(r"(", r"\(").replace(r")", r"\)")

        # QName cannot end with .
        if local.endswith("."):
            return None

        prefix = self.addNamespace(prefix, namespace)

        return "%s:%s" % (prefix, local)

    def startDocument(self) -> None:
        self._started = True
        ns_list = sorted(self.namespaces.items())

        if self.base:
            self.write(self.indent() + "@base <%s> .\n" % self.base)
        for prefix, uri in ns_list:
            self.write(self.indent() + "@prefix %s: <%s> .\n" % (prefix, uri))
        if ns_list and self._spacious:
            self.write("\n")

    def endDocument(self) -> None:
        if self._spacious:
            self.write("\n")

    def statement(self, subject: _SubjectType) -> bool:
        self.subjectDone(subject)
        return self.s_squared(subject) or self.s_default(subject)

    def s_default(self, subject: _SubjectType) -> bool:
        self.write("\n" + self.indent())
        self.path(subject, SUBJECT)
        self.predicateList(subject)
        self.write(" .")
        return True

    def s_squared(self, subject: _SubjectType) -> bool:
        if (self._references[subject] > 0) or not isinstance(subject, BNode):
            return False
        self.write("\n" + self.indent() + "[]")
        self.predicateList(subject)
        self.write(" .")
        return True

    def path(self, node: Node, position: int, newline: bool = False) -> None:
        if not (
            self.p_squared(node, position, newline)
            or self.p_default(node, position, newline)
        ):
            raise Error("Cannot serialize node '%s'" % (node,))

    def p_default(self, node: Node, position: int, newline: bool = False) -> bool:
        if position != SUBJECT and not newline:
            self.write(" ")
        self.write(self.label(node, position))
        return True

    def label(self, node: Node, position: int) -> str:
        if node == RDF.nil:
            return "()"
        if position is VERB and node in self.keywords:
            return self.keywords[node]
        if isinstance(node, Literal):
            return node._literal_n3(
                use_plain=True,
                qname_callback=lambda dt: self.getQName(dt, _GEN_QNAME_FOR_DT),
            )
        else:
            node = self.relativize(node)  # type: ignore[type-var]

            return self.getQName(node, position == VERB) or node.n3()

    def p_squared(self, node: Node, position: int, newline: bool = False) -> bool:
        if (
            not isinstance(node, BNode)
            or node in self._serialized
            or self._references[node] > 1
            or position == SUBJECT
        ):
            return False

        if not newline:
            self.write(" ")

        if self.isValidList(node):
            # this is a list
            self.write("(")
            self.depth += 1  # 2
            self.doList(node)
            self.depth -= 1  # 2
            self.write(" )")
        else:
            self.subjectDone(node)
            self.depth += 2
            # self.write('[\n' + self.indent())
            self.write("[")
            self.depth -= 1
            # self.predicateList(node, newline=True)
            self.predicateList(node, newline=False)
            # self.write('\n' + self.indent() + ']')
            self.write(" ]")
            self.depth -= 1

        return True

    def isValidList(self, l_: Node) -> bool:
        """
        Checks if l is a valid RDF list, i.e. no nodes have other properties.
        """
        try:
            if self.store.value(l_, RDF.first) is None:
                return False
        except Exception:
            return False
        while l_:
            if l_ != RDF.nil and len(list(self.store.predicate_objects(l_))) != 2:
                return False
            # type error: Incompatible types in assignment (expression has type "Optional[Node]", variable has type "Node")
            l_ = self.store.value(l_, RDF.rest)  # type: ignore[assignment]
        return True

    def doList(self, l_: Node) -> None:
        while l_:
            item = self.store.value(l_, RDF.first)
            if item is not None:
                self.path(item, OBJECT)
                self.subjectDone(l_)
            # type error: Incompatible types in assignment (expression has type "Optional[Node]", variable has type "Node")
            l_ = self.store.value(l_, RDF.rest)  # type: ignore[assignment]

    def predicateList(self, subject: Node, newline: bool = False) -> None:
        properties = self.buildPredicateHash(subject)
        propList = self.sortProperties(properties)
        if len(propList) == 0:
            return
        self.verb(propList[0], newline=newline)
        self.objectList(properties[propList[0]])
        for predicate in propList[1:]:
            self.write(" ;\n" + self.indent(1))
            self.verb(predicate, newline=True)
            self.objectList(properties[predicate])

    def verb(self, node: Node, newline: bool = False) -> None:
        self.path(node, VERB, newline)

    def objectList(self, objects: Sequence[Node]) -> None:
        count = len(objects)
        if count == 0:
            return
        depthmod = (count == 1) and 0 or 1
        self.depth += depthmod
        self.path(objects[0], OBJECT)
        for obj in objects[1:]:
            self.write(",\n" + self.indent(1))
            self.path(obj, OBJECT, newline=True)
        self.depth -= depthmod



from unittest import mock

import pytest

import rdflib
from rdflib import ConjunctiveGraph

QUERY = """
SELECT DISTINCT ?g
FROM NAMED <http://ns.example.com/named#>
WHERE {
  GRAPH ?g {
    ?s ?p ?o .
  }
}
"""


def test_named_graph_with_fragment():
    """Test that fragment part of the URL is not erased."""
    graph = ConjunctiveGraph()

    with mock.patch("rdflib.parser.URLInputSource") as load_mock:
        # We have to expect an exception here.
        with pytest.raises(Exception):
            graph.query(QUERY)

    load_mock.assert_called_with(
        rdflib.URIRef("http://ns.example.com/named#"),
        "nt",
    )



import xml.dom.minidom
from typing import Callable

import pytest

import rdflib.term
from rdflib.namespace import RDF
from rdflib.term import Literal
from test.utils.literal import LiteralChecker
from test.utils.outcome import OutcomeChecker, OutcomePrimitives

try:
    import html5rdf as _  # noqa: F401
except ImportError:
    pytest.skip("html5rdf not installed", allow_module_level=True)


def test_has_html5rdf() -> None:
    assert rdflib.term._HAS_HTML5RDF is True
    assert RDF.HTML in rdflib.term.XSDToPython
    rule = next(
        (
            item
            for item in rdflib.term._GenericPythonToXSDRules
            if item[0] is xml.dom.minidom.DocumentFragment
        ),
        None,
    )
    assert rule is not None
    assert rule[1][1] == RDF.HTML


@pytest.mark.parametrize(
    ["factory", "outcome"],
    [
        # Ill-typed literals, these have lexical forms that result in
        # errors when parsed as HTML by html5rdf.
        (
            lambda: Literal("<body><h1>Hello, World!</h1></body>", datatype=RDF.HTML),
            LiteralChecker(
                ..., None, RDF.HTML, True, "<body><h1>Hello, World!</h1></body>"
            ),
        ),
        (
            lambda: Literal("<body></body>", datatype=RDF.HTML),
            LiteralChecker(..., None, RDF.HTML, True, "<body></body>"),
        ),
        (
            lambda: Literal("<tr><td>THE TEXT IS IN HERE</td></tr>", datatype=RDF.HTML),
            LiteralChecker(
                ..., None, RDF.HTML, True, "<tr><td>THE TEXT IS IN HERE</td></tr>"
            ),
        ),
        # Well-typed literals, these have lexical forms that parse
        # without errors with html5rdf.
        (
            lambda: Literal("<table></table>", datatype=RDF.HTML),
            LiteralChecker(..., None, RDF.HTML, False, "<table></table>"),
        ),
        (
            lambda: Literal("  <table>  </table>  ", datatype=RDF.HTML, normalize=True),
            LiteralChecker(..., None, RDF.HTML, False, "  <table>  </table>  "),
        ),
        (
            lambda: Literal(
                "  <table>  </table>  ", datatype=RDF.HTML, normalize=False
            ),
            LiteralChecker(..., None, RDF.HTML, False, "  <table>  </table>  "),
        ),
    ],
)
def test_literal_construction(
    factory: Callable[[], Literal],
    outcome: OutcomePrimitives[Literal],
) -> None:
    checker = OutcomeChecker[Literal].from_primitives(outcome)
    with checker.context():
        actual_outcome = factory()
        checker.check(actual_outcome)



import json

from rdflib import BNode, ConjunctiveGraph
from rdflib.compare import isomorphic
from rdflib.parser import InputSource
from rdflib.plugins.parsers.jsonld import JsonLDParser, to_rdf

# monkey-patch N-Quads parser via it's underlying W3CNTriplesParser to keep source bnode id:s ..
from rdflib.plugins.parsers.ntriples import W3CNTriplesParser, r_nodeid
from rdflib.plugins.serializers.jsonld import from_rdf
from rdflib.plugins.shared.jsonld.keys import CONTEXT, GRAPH


def _preserving_nodeid(self, bnode_context=None):
    if not self.peek("_"):
        return False
    return BNode(self.eat(r_nodeid).group(1))


DEFAULT_PARSER_VERSION = 1.0


def make_fake_urlinputsource(input_uri, format=None, suite_base=None, options={}):
    local_url = input_uri.replace("https://w3c.github.io/json-ld-api/tests/", "./")
    if (index := local_url.find("#")) > -1:
        # Strip off the optional fragment identifier
        local_url = local_url[0:index]
    try:
        f = open(local_url, "rb")
    except FileNotFoundError:
        f = None
    source = InputSource(input_uri)
    source.setPublicId(input_uri)
    source.setByteStream(f)
    source.url = input_uri
    source.links = []
    if local_url.endswith((".jsonld", ".jldt")):
        source.content_type = "application/ld+json"
    elif local_url.endswith(".html"):
        source.content_type = "text/html"
    else:
        source.content_type = "application/json"
    source.format = format
    if options:
        if "httpLink" in options:
            source.links.append(options["httpLink"])
        if "contentType" in options:
            source.content_type = options["contentType"]
        if "redirectTo" in options:
            redir = suite_base + options["redirectTo"]
            local_redirect = redir.replace(
                "https://w3c.github.io/json-ld-api/tests/", "./"
            )
            if f:
                f.close()
            try:
                f = open(local_redirect, "rb")
            except FileNotFoundError:
                f = None
            source.setByteStream(f)
            source.url = redir
            source.setPublicId(redir)
            source.setSystemId(redir)
    return source


def do_test_json(suite_base, cat, num, inputpath, expectedpath, context, options):
    input_uri = suite_base + inputpath
    input_graph = ConjunctiveGraph()
    if cat == "remote-doc":
        input_src = make_fake_urlinputsource(
            input_uri, format="json-ld", suite_base=suite_base, options=options
        )
        p = JsonLDParser()
        p.parse(
            input_src,
            input_graph,
            base=input_src.getPublicId(),
            context_data=context,
            generalized_rdf=True,
        )
    else:
        input_obj = _load_json(inputpath)
        to_rdf(
            input_obj,
            input_graph,
            base=input_uri,
            context_data=context,
            generalized_rdf=True,
        )
    expected_json = _load_json(expectedpath)
    use_native_types = True  # CONTEXT in input_obj
    result_json = from_rdf(
        input_graph,
        context,
        base="./",  # deliberately set base different to the input base
        use_native_types=options.get("useNativeTypes", use_native_types),
        use_rdf_type=options.get("useRdfType", False),
    )

    def _prune_json(data):
        if CONTEXT in data:
            data.pop(CONTEXT)
        if GRAPH in data:
            data = data[GRAPH]
        # def _remove_empty_sets(obj):
        return data

    expected_json = _prune_json(expected_json)
    result_json = _prune_json(result_json)

    _compare_json(expected_json, result_json)


def do_test_parser(suite_base, cat, num, inputpath, expectedpath, context, options):
    input_uri = suite_base + inputpath
    input_obj = _load_json(inputpath)
    old_nodeid = W3CNTriplesParser.nodeid
    # monkey patch nodeid fn in NTriplesParser
    W3CNTriplesParser.nodeid = _preserving_nodeid
    try:
        expected_graph = _load_nquads(expectedpath)
    finally:
        W3CNTriplesParser.nodeid = old_nodeid
    result_graph = ConjunctiveGraph()
    requested_version = options.get("specVersion")
    version = DEFAULT_PARSER_VERSION
    if requested_version:
        if requested_version == "json-ld-1.1":
            version = 1.1
        elif requested_version == "json-ld-1.0":
            version = 1.0
    if cat == "remote-doc":
        input_src = make_fake_urlinputsource(
            input_uri, format="json-ld", options=options
        )
        p = JsonLDParser()
        p.parse(
            input_src,
            result_graph,
            base=input_uri,
            context_data=context,
            generalized_rdf=True,
        )
    else:
        to_rdf(
            input_obj,
            result_graph,
            context_data=context,
            base=options.get("base", input_uri),
            version=version,
            generalized_rdf=options.get("produceGeneralizedRdf", False),
        )
    assert isomorphic(result_graph, expected_graph), "Expected:\n%s\nGot:\n%s" % (
        expected_graph.serialize(),
        result_graph.serialize(),
    )


def do_test_serializer(suite_base, cat, num, inputpath, expectedpath, context, options):
    input_uri = suite_base + inputpath
    old_nodeid = W3CNTriplesParser.nodeid
    # monkey patch nodeid fn in NTriplesParser
    W3CNTriplesParser.nodeid = _preserving_nodeid
    try:
        input_graph = _load_nquads(inputpath)
    finally:
        W3CNTriplesParser.nodeid = old_nodeid
    expected_json = _load_json(expectedpath)
    result_json = from_rdf(
        input_graph,
        context,
        base=input_uri,
        use_native_types=options.get("useNativeTypes", False),
        use_rdf_type=options.get("useRdfType", False),
    )
    _compare_json(expected_json, result_json)


def do_test_html(suite_base, cat, num, inputpath, expectedpath, context, options):
    input_uri = suite_base + inputpath
    input_graph = ConjunctiveGraph()

    input_src = make_fake_urlinputsource(
        input_uri, format="json-ld", suite_base=suite_base, options=options
    )

    context = _load_json(context) if context else context

    # Get test options from the manifest
    base = options.get("base", input_src.getPublicId())
    extract_all_scripts = options.get("extractAllScripts", False)

    p = JsonLDParser()
    p.parse(
        input_src,
        input_graph,
        base=base,
        context=context,
        generalized_rdf=True,
        extract_all_scripts=extract_all_scripts,
    )

    if expectedpath.endswith(".nq"):
        expected_graph = _load_nquads(expectedpath)

    elif expectedpath.endswith(".jsonld"):
        expected_graph = ConjunctiveGraph()
        with open(expectedpath) as f:
            data = f.read()
        expected_graph.parse(data=data, format="json-ld")

    # TODO: Change test from graph comparison to json comparison
    # The html test cases combine testing for JSON-LD extraction from the HTML
    # along with testing for other algorithms (compact/flatten), which we do
    # not currently support. In order to test extraction only, we currently
    # perform a graph comparison. Consider changing this to a json comparison
    # once the processing algorithms are implemented.

    assert isomorphic(input_graph, expected_graph), "Expected:\n%s\nGot:\n%s" % (
        expected_graph.serialize(),
        input_graph.serialize(),
    )


def _load_nquads(source):
    graph = ConjunctiveGraph()
    with open(source) as f:
        data = f.read()
    graph.parse(data=data, format="nquads")
    return graph


def _load_json(source):
    with open(source) as f:
        return json.load(f)


def _to_ordered(obj):
    if isinstance(obj, list):
        # NOTE: use type in key to handle mixed
        # lists of e.g. bool, int, float.
        return sorted(
            (_to_ordered(lv) for lv in obj),
            key=lambda x: (_ord_key(x), type(x).__name__),
        )
    if not isinstance(obj, dict):
        return obj
    return sorted((k, _to_ordered(v)) for k, v in obj.items())


def _ord_key(x):
    if isinstance(x, dict) and "@id" in x:
        return x["@id"]
    else:
        return x


def _dump_json(obj):
    return json.dumps(
        obj, indent=2, separators=(",", ": "), sort_keys=True, check_circular=True
    )


def _compare_json(expected, result):
    expected = json.loads(_dump_json(expected))
    result = json.loads(_dump_json(result))
    assert _to_ordered(expected) == _to_ordered(
        result
    ), "Expected JSON:\n%s\nGot:\n%s" % (_dump_json(expected), _dump_json(result))



"""
Issue 715 - path query chaining issue
Some incorrect matches were found when using oneOrMore ('+') and
zeroOrMore ('*') property paths and specifying neither the
subject or the object.
"""

from rdflib import Graph
from test.utils.namespace import EGDO


def test_issue_733():
    g = Graph()
    g.add((EGDO.S, EGDO.P, EGDO.O1))
    g.add((EGDO.S, EGDO.P, EGDO.O2))
    q = """
    prefix ex:<http://example.org/>
    select ?lexical_or_value ?ot ?gt where {
        {SELECT (count(*) as ?lexical_or_value) where {
        ?s ?p ?o .
            FILTER (?s=ex:S)
        }}
        {SELECT (count(*) as ?ot) where {
        ?s ?p ?o .
            FILTER (?o=ex:O1)
        }}
        {SELECT (count(*) as ?gt) where {
        ?s ?p ?o .
            FILTER (?o!=ex:O1 && ?s!=ex:O2)
        }}
    }
    """
    res = g.query(q)
    assert len(res) == 1
    results = [[lit.toPython() for lit in line] for line in res]
    assert results[0][0] == 2
    assert results[0][1] == 1
    assert results[0][2] == 1


def test_issue_733_independant():
    g = Graph()
    g.add((EGDO.S, EGDO.P, EGDO.O1))
    g.add((EGDO.S, EGDO.P, EGDO.O2))
    q = """
            prefix ex:<http://example.org/>
            select ?lexical_or_value where {
                {SELECT (count(*) as ?lexical_or_value) where {
                ?s ?p ?o .
                    FILTER (?s=ex:S)
                }}
            }
            """
    res = g.query(q)
    assert len(res) == 1
    results = [[lit.toPython() for lit in line] for line in res]
    assert results[0][0] == 2
    q = """
            prefix ex:<http://example.org/>
            select ?lexical_or_value where {
                {SELECT (count(*) as ?lexical_or_value) where {
                ?s ?p ?o .
                    FILTER (?o=ex:O1)
                }}
            }
            """
    res = g.query(q)
    results = [[lit.toPython() for lit in line] for line in res]
    assert results[0][0] == 1
